#include <stdio.h>
#include <signal.h>
#include <sys/time.h>

#define INTERVAL 999999
#define QUEUE_SZ 20
#define MAX_CMD_LENGTH 100

#define SUCC(x) ((x + 1) % QUEUE_SZ)

typedef void (*sighandler_t)(int);
typedef struct _queue {
  int head;
  int tail;
  char buffer[QUEUE_SZ][MAX_CMD_LENGTH];
} queue;

queue *make_queue(void);
void add_cmd(const char *cmd, queue *q);
char *get_cmd(queue *q);

static struct itimerval step = { { 0, INTERVAL }, { 0, INTERVAL }};

queue* make_queue(void) {
  /* Qui devi mettere il codice che restituisce il necessario per
     essere passato al LISP come stringa */
  queue *ret = (queue*)malloc(sizeof(queue));
  ret->head = ret->tail = 0;
  
  return ret;
}

void add_cmd(const char *cmd, queue *q) {
  int tail = SUCC(q->tail);
  if (tail != q->head) {
    strcpy(q->buffer[q->tail], cmd);
    q->tail = tail;
  }
}

char* get_cmd(queue *q) {
  int head = q->head;

  if (q->tail == head) return NULL;
  
  q->head = SUCC(q->head);
  return q->buffer[head];
}

void fire(int type);

static int count = 1;

void fire(int type) {
  while (setitimer(ITIMER_REAL, &step, NULL) == -1);

  printf("%i: Signal received!\n", count++);

  if (signal(SIGALRM, (sighandler_t)fire) == SIG_ERR) {
    fprintf(stderr, "Error installing the signal!");
    exit(1);
  }
}

int main(int argc, char **argv) {
  int i;
  struct itimerval timer;

  queue* q = make_queue();
  add_cmd("Prova 1", q);
  add_cmd("Prova 2", q);
  puts(get_cmd(q));
  add_cmd("Prova 3", q);
  puts(get_cmd(q));
  puts(get_cmd(q));
  if (!get_cmd(q)) puts("Empty queue");
  else puts("Error!");
  
  if (signal(SIGALRM, (sighandler_t)fire) == SIG_ERR) {
    fprintf(stderr, "Error installing the signal!");
    exit(1);
  }

  if  (setitimer(ITIMER_REAL, &step, NULL) == -1) exit(1);

  puts("Waiting 25 seconds...");
  for (i = 0;i < 1000000000; i++);
  puts("Done.");

  return 0;
}





