;;;; robocup.lsp
;;;; Last update: 28/9/98
;;;; @author Antonio Cisternino
;;;; @version 1.1
;;;; Changes:
;;;;
;;;; 28/9/98: Removed the global variable *client-udp*. The variable
;;;;         is keep into the C code.
;;;;
;;;; 28/9/98: Added the queue support and the signal support to send
;;;;          messages in an asynchronous way to the soccerserver

;;; Header del file C generato dal compilatore
(clines "
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <sys/time.h>
#include <signal.h>

#define INTERVAL 100000
#define QUEUE_SZ 20
#define MAX_CMD_LENGTH 100
#define SUCC(x) ((x + 1) % QUEUE_SZ)

typedef void (*sighandler_t)(int);

static struct socket {
    int			socketfd;
    struct sockaddr_in	serv_addr;
} _sock, *sock = &_sock;

static struct rc_queue {
  int head;
  int tail;
  char buffer[QUEUE_SZ][MAX_CMD_LENGTH];
} _q, *q = &_q;

static struct itimerval step = { { 0, INTERVAL }, { 0, INTERVAL }};

void add_cmd(const char *cmd) {
  int tail = SUCC(q->tail);

  if (tail != q->head) {
    strcpy(q->buffer[q->tail], cmd);
    q->tail = tail;
  }
}

char* get_cmd() {
  int head = q->head;

  if (q->tail == head) return NULL;

  q->head = SUCC(q->head);
  return q->buffer[head];
}

void fire(int type) {
  char *cmd;
  int n;

  while (setitimer(ITIMER_REAL, &step, NULL) == -1);
  if (signal(SIGALRM, (sighandler_t)fire) == SIG_ERR) {
    fprintf(stderr, \"Error installing the signal!\");
    exit(1);
  }

  cmd = get_cmd();

  if (!cmd) return;

  n = strlen(cmd);

  if (sendto(sock->socketfd, cmd, n, 0,
	        (struct sockaddr *)&sock->serv_addr, sizeof(sock->serv_addr))
	  != n) {
    fprintf(stderr, \"Error sending data!\");
    exit(1);
  }
}
")

;;; Si connette al server
;;; @param object La stringa contenente il nome del server
;;; @param object La stringa contenente la porta a cui connettersi
(defcbody connect-server-soccer (object object) object
"
  Cnil;
  {
    struct hostent 	*host_ent;
    struct in_addr      *addr_ptr;
    struct sockaddr_in	cli_addr;
    int			sockfd, val;

    int serverPort = atoi(#1->st.st_self);

    if (signal(SIGALRM, (sighandler_t)fire) == SIG_ERR) {
      fprintf(stderr, \"Error installing the signal!\");
      exit(1);
    }

    if  (setitimer(ITIMER_REAL, &step, NULL) == -1) exit(1);

    q->tail = q->head = 0;
    sock->socketfd = -1;

    if ((host_ent = (struct hostent *)gethostbyname(#0->st.st_self)) == NULL) {
	/* Check if a numeric address */
	if (inet_addr(#0->st.st_self) == -1) {
	    VALUES(0) = Cnil;
        }
    } else {
	addr_ptr = (struct in_addr *) *host_ent->h_addr_list;
	#0->st.st_self = inet_ntoa(*addr_ptr);
    }


    /*
     *  Open UDP socket.
     */
    if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	VALUES(0) = Cnil;  	/* Can't open socket. */
    }

    val = fcntl(sockfd, F_GETFL, 0);
    val |= O_NONBLOCK;
    fcntl(sockfd, F_SETFL, val);

    /*
     *  Bind any local address.
     */
    memset((char *) &cli_addr, 0, sizeof(cli_addr));
    cli_addr.sin_family		= AF_INET;
    cli_addr.sin_addr.s_addr	= htonl(INADDR_ANY);
    cli_addr.sin_port		= htons(0);

    if (bind(sockfd, (struct sockaddr *) &cli_addr,
	    sizeof(cli_addr)) < 0) {
	VALUES(0) = Cnil;  /* Can't bind local address */
    }

    /*
     *  Fill in the structure with the address of the server.
     */
    sock->socketfd = sockfd;

    memset((char *) &sock->serv_addr, 0, sizeof(sock->serv_addr));
    sock->serv_addr.sin_family		= AF_INET;
    sock->serv_addr.sin_addr.s_addr	= inet_addr(#0->st.st_self);
    sock->serv_addr.sin_port		= htons(serverPort);

    VALUES(0) = Ct;

    RETURN(1);
  } 
"
)

;;; Invia un messaggio al server
;;; @param object Stringa da inviare al server
;;; @return T se il pacchetto e' stato inviato nil altrimenti.
(defcbody send-msg (object) object
"
  Cnil;
  {
    add_cmd(#0->st.st_self);

    VALUES(0) = Ct;

    RETURN(1);
  }
"
)

;;; Riceve un pacchetto dal server
;;; @return La stringa contenente il messaggio
(defcbody receive-msg () object
"
  Cnil;
  {
    int			n,servlen;
    struct sockaddr_in	serv_addr;
    char buf[2048];

    object ret;
    
    servlen = sizeof(serv_addr);

    do {
      n = recvfrom(sock->socketfd, buf, 2048, 0,
			 (struct sockaddr *)&serv_addr, &servlen);
    } while (n == -1); /* blocca */

    sock->serv_addr.sin_port = serv_addr.sin_port;
    buf[n] = '\\0';
    ret = (object)make_string_copy(buf);
    VALUES(0) = ret;

    RETURN(1);
  }
"
)

;;; Riceve l'ultimo messaggio inviato dal server ignorando tutti i precedenti
;;; @return La stringa contenente il messaggio
(defcbody receive-last-msg () object
"
  Cnil;
  {
    int			n,servlen;
    struct sockaddr_in	serv_addr;
    char buf[2048], retstr[2048];

    object ret;

    servlen = sizeof(serv_addr);

    retstr[0] = '\\0';

    do {
      n = recvfrom(sock->socketfd, buf, 2048, 0,
			 (struct sockaddr *)&serv_addr, &servlen);
      buf[n] = '\\0';
      if (n != -1) strcpy(retstr, buf);

    } while(!retstr[0] || (n != -1)); /* blocca */

    if (errno != EWOULDBLOCK) {
	VALUES(0) = Cnil;
    } else {
      sock->serv_addr.sin_port = serv_addr.sin_port;
      retstr[n] = '\\0';
      ret = (object)make_string_copy(retstr);
      VALUES(0) = ret;
    }

    RETURN(1);
  }
"
)

;;; Chiude la connessione col server
;;; @return nil
(defcbody close-connection () object
"
  Cnil;
  {
    close(sock->socketfd) ;

    VALUES(0) = Cnil;

    RETURN(1);
  }
"
)


;;; Queste funzioni elevano il livello della connessione con il lisp

(defun connect-to-server (host &optional (port 6000))
  "Si collega al server
  @param host Stringa contenente il nome della macchina su cui gira il server
  @param port Porta a cui connettersi. Se omessa viene assunto il valore 6000
  @return T se la connessione e' avvenuta correttamente NIL altrimenti"
  (if (and (< port 65536) (> port 0))
      (if (connect-server-soccer host (princ-to-string port))
	  T
	NIL)
    NIL))

(defun read-percept ()
  "Legge un messaggio dal server e lo converte in una S-Expr
  @return L'S-Expr o NIL se qualcosa e' andato male"
  (let ((msg (receive-msg)))
    (if msg
	(read-from-string msg)
      NIL)))

(defun read-last-percept()
  "legge l'ultimo messaggio inviato dal server, ignorando i precedenti,
   e restituisce l's-expr
   @return L'S-Expr o NIL se qualcosa e' andato male"
  (let ((msg (receive-last-msg)))
    (if msg
	(read-from-string msg)
      NIL)))

(defun do-action (action-list)
"Invia una s-expr al server
  @param action S-Expr da inviare al server
  @return T se l'S-Expr e' stato inviato nil altrimenti."
(dolist (action action-list)
  (send-msg (string-downcase (princ-to-string action)))))

(defun end-connection ()
"Chiude la connessione col server
  @return NIL"
  (close-connection))

