;;;; Allenatore che permette interattivamente di costruire situazioni
;;;; nel campo. Supporta la registrazione di macro per ripetere situazioni.
;;;; Last update: 30/12/1998
;;;; @author Maria Simi
;;;; @author Antonio Cisternino
;;;; @version 0.2
;;;;
;;;; Changes:
;;;;
;;;; 12/4/1999: Now the coach has become part of the library distribution
;;;;
;;;; 30/12/1998: Changed the location of robocup.o. Now the code searches
;;;;             for ../lib/robocup-2.0.o


(defun coach (&optional (host "localhost"))
  "Avvia una sessione di coach interattivo. Il comando '?' al prompt dei comandi informa sui comandi disponibili."
  (let (act (macros (make-hash-table)) (record nil))
    (connect-to-server host 6001)
    (loop
     (if record
	 (format t "coach@~A*~A> " host record)
       (format t "coach@~A> " host))
     (setf act (read))
     (cond

      ;; Give some help
      ((eql act '?)
       (princ "
 Available commands:

   (* <seq name>) - Start the record of a sequence
   (H)            - Ends the record
   (> <seq name>) - Plays the given sequence
   q              - Breaks the coach loop
   s              - Saves the current macros to a file
   l              - Loads the macros from file

 Other commands are sent to the soccer server.

"))

      ;; Quit the coach
      ((eql act 'q)
       (return 'Done))

      ;; Saves data to a file
      ((eql act 's)
       (let (fname)
	 (princ "Output file name: ") ; request the file name
	 (setf fname (read-line))
	 (when (probe-file fname)     ; prompt for overriding
	   (format t "~A already exists do you want override (Y/N)? " fname)
	   (unless (eql (read) 'y)
	     (setf fname nil)))
	 (when fname                  ; save the file
	   (with-open-file (out fname :direction :output :if-exists :supersede)
			   (maphash #'(lambda (x y) 
					(when y
					  (format out "(~A ~A)~%" x y)))
				    macros))
	   (format t "Saved ~A.~%" fname))))

      ;; Loads data from file
      ((eql act 'l)
       (let (fname)
	 (princ "Input file name: ") ; request the file name
	 (setf fname (read-line))
	 (unless (probe-file fname)  ; check for fname
	   (format t "Cannot find ~A!~%" fname)
	   (setf fname nil))
	 (when fname                 ; load the file
	   (clrhash macros)
	   (with-open-file (in fname :direction :input)
			   (do ((el (read in nil nil) (read in nil nil)))
			       ((not el))
			       (setf (gethash (car el) macros) (cadr el))))
	   (format t "~A loaded.~%" fname))))

      ;; Start recording a sequence with a given name
      ((and (consp act) (eql (car act) '*))
       (setf record (cadr act))
       (if (gethash record macros)
	 (loop
	  (format t "The ~A sequence already exists, do you want A)ppend or R)ewrite it? " record)
	  (case (read)
		('A (return nil))
		('R (setf (gethash record macros) nil) (return nil))))
	 (setf (gethash record macros) nil))
       (format t "Recording sequence ~A~%" record))

      ;; Stop the recording of a sequence
      ((and (consp act) (eql (car act) 'H))
       (format t "Stopping recording ~A sequence~%" record)
       (setf record nil))

      ;; Plays a recorded sequence
      ((and (consp act) (eql (car act) '>))
       (if (gethash (cadr act) macros)
	   (dolist (el (gethash (cadr act) macros))
		   (format t "Executing: ~A~%" el)
		   (do-action el)
		   (format t "Result:    ~A~%" (read-percept)))
	 (format t "Sequence ~A is unknown!~%" (cadr act))))

      ;; Default: sends the data to the server
      (t
       (when record 
	 (setf (gethash record macros)
	       (append (gethash record macros) (list act))))
       (do-action act)
       (prin1 (read-percept))
       (terpri))))))
