/****************************************************************

  Author: Antonio Cisternino

  Version: 1.5.24

  Description: This file contains some type definition.

 ****************************************************************/

#ifndef __PARSER_H__
#define __PARSER_H__

#define MAX_TEAMNAME_LEN 20
#define MAX_MSG_LEN      512
#define INVALID_VALUE    65536
#define SELF_DIR         65537
#define REFEREE_DIR      65538

typedef enum { OK, INTERNAL_ERROR, UNKNOWN_PERCEPT } rc_result_t;
typedef enum { OBJ_PLAYER, OBJ_GOAL, OBJ_BALL, OBJ_LINE, OBJ_FLAG } rc_obj_t;
typedef enum { S_UNKNOWN, S_BEFORE_KICK_OFF, S_KICK_OFF, S_KICK_IN,
	       S_CORNER_KICK, S_OFFSIDE, S_GOAL_KICK, S_FREE_KICK, 
	       S_PLAY_ON, S_HALF_TIME, S_TIME_UP, S_EXTEND, S_FOUL,
	       S_GOAL } rc_state_t;
typedef enum { VQ_HIGH, VQ_LOW } rc_view_quality_t;
typedef enum { VW_NARROW, VW_NORMAL, VW_WIDE } rc_view_width_t;
typedef enum { E_UNKNOWN_COMMAND, E_ILLIGAL_COMMAND_FORM } rc_error_t;

rc_result_t analyze(char *p);

#define READ_NUM(dest) { \
	int afterdot = 0, power = 1, sign = 1; \
        dest = 0; \
\
	if (*input == '-') { \
	  sign = -1; \
	  ++input; \
	} \
\
	while (isdigit(*input) || (*input == '.')) { \
	  if (*input == '.') \
	    afterdot = 1; \
	  else { \
	    if (afterdot) power *= 10; \
	    dest = (dest * 10) + (*input - '0'); \
	  } \
\
	  ++input; \
	} \
\
        dest *= sign; \
	if (power > 1) \
	  dest /= power; \
      }

#define READ_INT(dest) dest = 0; \
  while (isdigit(*input)) \
    dest = (dest * 10) + (*input++ - '0');

#define CHECK_STATE(WHERE)   switch (*input) { \
  case 'b': /* before_kick_off */ \
    state = S_BEFORE_KICK_OFF; \
    break; \
  case 'k': /* kick_* */ \
    if (input[5] == 'i') { \
      state = S_KICK_IN; \
      WHERE = input[8]; \
    } else { \
      state = S_KICK_OFF; \
      WHERE = input[9]; \
    } \
    break; \
  case 'c': /* corner_kick */ \
    state = S_CORNER_KICK; \
    WHERE = input[12]; \
    break; \
  case 'o': /* offside */ \
    state = S_OFFSIDE; \
    WHERE = input[8]; \
    break; \
  case 'g': /* goal_* */ \
    if (input[5] == 'k') { \
      state = S_GOAL_KICK; \
      WHERE = input[10]; \
    } else { \
      state = S_GOAL; \
      WHERE = input[5]; \
      input += 7; \
      READ_INT(num); \
    } \
    break; \
  case 'f': /* f* */ \
    if (input[1] == 'r') { \
      state = S_FREE_KICK; \
      WHERE = input[10]; \
    } else { \
      state = S_FOUL; \
      WHERE = input[5]; \
      input += 7; \
      READ_INT(num); \
    } \
    break; \
  case 'p': /* play_on */ \
    state = S_PLAY_ON; \
    break; \
  case 'h': /* half_time */ \
    state = S_HALF_TIME; \
    break; \
  case 't': /* time_up */ \
    state = S_TIME_UP; \
    break; \
  case 'e': /* extend */ \
    state = S_EXTEND; \
    break; \
  default: \
    state = S_UNKNOWN; \
  }


#endif /* __PARSER_H__ */
