#ifndef __QUEUE_H__
#define __QUEUE_H__

#include "smartlow.h"

/** Size of the commands' queue */
#define QUEUE_SZ 6
/** Calculate the successor index in the circular queue */
#define SUCC(x) ((x + 1) % QUEUE_SZ)

/** This structure defines the queue */
typedef struct {
  int head;
  int tail;
  char buffer[QUEUE_SZ][MAX_CMD_LENGTH];
} smart_queue;

/**
 * Discard all the commands in the queue and set the queue's length
 * to zero.
 */
void smart_reset_queue();

/**
 * Add a command to the queue. The command is represented by a
 * common string. In this function is implemented the turn
 * trashing.
 * @param cmd Command that must be enqueued
 * @see smart_turn_trashing(int)
 */
void smart_add_queue(const char *cmd);

/**
 * Get a command from the queue.
 * @return the first command in the queue or NULL if the queue is empty.
 */
char* smart_get_queue();

/**
 * Return the length of the queue.
 * @return The length of the queue.
 */
int smart_queue_length();

#endif /* __QUEUE_H__ */
