;;;; robocup.lsp
;;;; Last update: 16/4/98
;;;; @author Antonio Cisternino
;;;; @version 1.0

;;; Header del file C generato dal compilatore
(clines "
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/socket.h>
#include <stdlib.h>

typedef struct socket{
    int			socketfd;
    struct sockaddr_in	serv_addr;
} Socket;
")

;;; Si connette al server
;;; @param object La stringa contenente il nome del server
;;; @param object La stringa contenente la porta a cui connettersi
;;; @return Il socket UDP
(defcbody connect-server-soccer (object object) object
"
  Cnil;
  {
    struct hostent 	*host_ent;
    struct in_addr      *addr_ptr;
    struct sockaddr_in	cli_addr;
    int			sockfd, val;
    
    object s = make_string_copy(\"01234567890123456789012345\");
    Socket *sock = (Socket*)s->st.st_self;
    int serverPort = atoi(#1->st.st_self);

    sock->socketfd = -1;

    if ((host_ent = (struct hostent *)gethostbyname(#0->st.st_self)) == NULL) {
	/* Check if a numeric address */
	if (inet_addr(#0->st.st_self) == -1) {
	    VALUES(0) = Cnil;
        }
    } else {
	addr_ptr = (struct in_addr *) *host_ent->h_addr_list;
	#0->st.st_self = inet_ntoa(*addr_ptr);
    }


    /*
     *  Open UDP socket.
     */
    if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	VALUES(0) = Cnil;  	/* Can't open socket. */
    }

    val = fcntl(sockfd, F_GETFL, 0);
    val |= O_NONBLOCK;
    fcntl(sockfd, F_SETFL, val);

    /*
     *  Bind any local address.
     */
    memset((char *) &cli_addr, 0, sizeof(cli_addr));
    cli_addr.sin_family		= AF_INET;
    cli_addr.sin_addr.s_addr	= htonl(INADDR_ANY);
    cli_addr.sin_port		= htons(0);

    if (bind(sockfd, (struct sockaddr *) &cli_addr,
	    sizeof(cli_addr)) < 0) {
	VALUES(0) = Cnil;  /* Can't bind local address */
    }

    /*
     *  Fill in the structure with the address of the server.
     */
    sock->socketfd = sockfd;

    memset((char *) &sock->serv_addr, 0, sizeof(sock->serv_addr));
    sock->serv_addr.sin_family		= AF_INET;
    sock->serv_addr.sin_addr.s_addr	= inet_addr(#0->st.st_self);
    sock->serv_addr.sin_port		= htons(serverPort);

    VALUES(0) = s;

    RETURN(1);
  } 
"
)

;;; Invia un messaggio al server
;;; @param object Socket restituito da connect-server-soccer
;;; @param object Stringa da inviare al server
;;; @return T se il pacchetto e' stato inviato nil altrimenti.
(defcbody send-msg (object object) object
"
  Cnil;
  {
    int n;
    Socket *sock = (Socket*)#0->st.st_self;

    n = strlen(#1->st.st_self);

    if (sendto(sock->socketfd, #1->st.st_self, n, 0,
	        (struct sockaddr *)&sock->serv_addr, sizeof(sock->serv_addr))
	  != n)
	VALUES(0) = Cnil;
    else
	VALUES(0) = Ct;

    RETURN(1);
  }
"
)

;;; Riceve un pacchetto dal server
;;; @param object Socket restituito da connect-server-soccer
;;; @return La stringa contenente il messaggio
(defcbody receive-msg (object) object
"
  Cnil;
  {
    int			n,servlen;
    struct sockaddr_in	serv_addr;
    Socket *sock = (Socket*)#0->st.st_self;
    char buf[2048];

    object ret;
    
    servlen = sizeof(serv_addr);

    do {
      n = recvfrom(sock->socketfd, buf, 2048, 0,
			 (struct sockaddr *)&serv_addr, &servlen);
    } while (n == -1); /* blocca */

    sock->serv_addr.sin_port = serv_addr.sin_port;
    buf[n] = '\\0';
    ret = make_string_copy(buf);
    VALUES(0) = ret;

    RETURN(1);
  }
"
)

;;; Riceve l'ultimo messaggio inviato dal server ignorando tutti i precedenti
;;; @param object Socket restituito da connect-server-soccer
;;; @return La stringa contenente il messaggio
(defcbody receive-last-msg (object) object
"
  Cnil;
  {
    int			n,servlen;
    struct sockaddr_in	serv_addr;
    Socket *sock = (Socket*)#0->st.st_self;
    char buf[2048], retstr[2048];

    object ret;

    servlen = sizeof(serv_addr);

    retstr[0] = '\\0';

    do {
      n = recvfrom(sock->socketfd, buf, 2048, 0,
			 (struct sockaddr *)&serv_addr, &servlen);
      buf[n] = '\\0';
      if (n != -1) strcpy(retstr, buf);

    } while(!retstr[0] || (n != -1)); /* blocca */

    if (errno != EWOULDBLOCK) {
	VALUES(0) = Cnil;
    } else {
      sock->serv_addr.sin_port = serv_addr.sin_port;
      retstr[n] = '\\0';
      ret = make_string_copy(retstr);
      VALUES(0) = ret;
    }

    RETURN(1);
  }
"
)

;;; Chiude la connessione col server
;;; @param object Socket restituito da connect-server-soccer
;;; @return nil
(defcbody close-connection (object) object
"
  Cnil;
  {
    Socket* sock = (Socket*)#0->st.st_self;
    close(sock->socketfd) ;

    VALUES(0) = Cnil;

    RETURN(1);
  }
"
)


;;; Queste funzioni elevano il livello della connessione con il lisp

(defvar *clientudp* NIL 
  "Variabile globale dove viene memorizzato il Socket UDP")

(defun connect-to-server (host &optional (port 6000))
"Si collega al server
  @param host Stringa contenente il nome della macchina su cui gira il server
  @param port Porta a cui connettersi. Se omessa viene assunto il valore 6000
  @return T se la connessione e' avvenuta correttamente NIL altrimenti"
  (if (and (< port 65536) (> port 0))
      (if (setf *clientudp* 
		(connect-server-soccer host (princ-to-string port)))
	  T
	NIL)
    NIL))

(defun read-percept ()
  "Legge un messaggio dal server e lo converte in una S-Expr
  @return L'S-Expr o NIL se qualcosa e' andato male"
  (let ((msg (receive-msg *clientudp*)))
    (if msg
	(read-from-string msg)
      NIL)))

(defun read-last-percept()
"legge l'ultimo messaggio inviato dal server, ignorando i precedenti,
e restituisce l's-expr
  @return L'S-Expr o NIL se qualcosa e' andato male"
  (let ((msg (receive-last-msg *clientudp*)))
    (if msg
	(read-from-string msg)
      NIL)))

(defun do-action (action)
"Invia una s-expr al server
  @param action S-Expr da inviare al server
  @return T se l'S-Expr e' stato inviato nil altrimenti."
  (send-msg *clientudp* (string-downcase (princ-to-string action))))

(defun end-connection ()
"Chiude la connessione col server
  @return NIL"
  (close-connection *clientudp*))

