;;;; robocup-2.0.lsp
;;;; Last update: 08/08/1999
;;;; @author Antonio Cisternino
;;;; @version 2.4
;;;;
;;;; Changes:
;;;;
;;;; 14/08/1999: Added the internal parser for messages sent by the server
;;;;
;;;; 08/08/1999: Added the dynamic timer control: at each cycle the timer
;;;;             is set to enhance the communication with the server.
;;;;
;;;; 06/08/1999: Now the player sleeps if no percept is found.
;;;;
;;;; 19/4/1999: Added the Sense body filtering
;;;;
;;;; 19/4/1999: Some new bug-fix on C code
;;;;
;;;; 6/4/1999: Added the rc-turn-trash-off and rc-turn-trash-on functions.
;;;;           These functions allow the control of queue filtering
;;;;           in response to 'turn' commands.
;;;;
;;;; 12/3/1999: The statistic about trashing has been introduced. If
;;;;            the read-last-percept is used instead of read-percept
;;;;            it is possible to use the rc-trashing to know how many
;;;;            messages have been trashed.
;;;;
;;;; 7/3/1999: The received last message has been changed: we stop
;;;;           when an hear message is found or the queue is finished
;;;;
;;;; 12/1/1999: When a command is inserted into the message queue and
;;;;            the queue is full the oldest action is discarded. This
;;;;            provides a sort of trashing control. If this is not
;;;;            sufficient I will try to change dinamically the size
;;;;            of the queue.
;;;;
;;;; 3/1/1999: If a turn action is submitted the queue is empty
;;;;           to syncronize the percept with the commands flow
;;;;           The C function 'add_cmd' contains the modify.
;;;;           Experimentation has shown that this check it is necessary
;;;;
;;;; 3/1/1999: INTERVAL has changed from 100000 to 90000
;;;; 
;;;; 3/1/1999: The QUEUE_SZ has changed to 6 from 20.
;;;;
;;;; 1/1/1999: Added a generic-code-loader to easy redefine
;;;;           a loader for the code contained in a given
;;;;           directory.
;;;;
;;;; 1/1/1999: Added some documentation
;;;;
;;;; 31/12/1998: Added the function rcload to load libraries from a
;;;;             standard directory.
;;;;
;;;; 30/12/1998: Added the method do-actions
;;;;
;;;; 28/9/1998: Removed the global variable *client-udp*. The variable
;;;;            is keep into the C code.
;;;;
;;;; 28/9/1998: Added the queue support and the signal support to send
;;;;            messages in an asynchronous way to the soccerserver

;;; Where libraries are stored
(defconstant *robocup-lib* "/home/cisterni/RoboCup/teams/IALP/lib")

;;; Main header
(clines "
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <sys/time.h>
#include <signal.h>
#include <ctype.h>

/* File parser.h included by the generator c2ecl */

#ifndef __PARSER_H__
#define __PARSER_H__

#define MAX_TEAMNAME_LEN 20
#define MAX_MSG_LEN      512
#define INVALID_VALUE    65536
#define SELF_DIR         65537
#define REFEREE_DIR      65538

#define READ_NUM(dest) { \\
	int afterdot = 0, power = 1, sign = 1; \\
        dest = 0; \\
\\
	if (*input == '-') { \\
	  sign = -1; \\
	  ++input; \\
	} \\
\\
	while (isdigit(*input) || (*input == '.')) { \\
	  if (*input == '.') \\
	    afterdot = 1; \\
	  else { \\
	    if (afterdot) power *= 10; \\
	    dest = (dest * 10) + (*input - '0'); \\
	  } \\
\\
	  ++input; \\
	} \\
\\
        dest *= sign; \\
	if (power > 1) \\
	  dest /= power; \\
      }

#define READ_INT(dest) dest = 0; \\
  while (isdigit(*input)) \\
    dest = (dest * 10) + (*input++ - '0');

typedef enum { OK, INTERNAL_ERROR, UNKNOWN_PERCEPT } rc_result_t;
typedef enum { OBJ_PLAYER, OBJ_GOAL, OBJ_BALL, OBJ_LINE, OBJ_FLAG } rc_obj_t;
typedef enum { S_UNKNOWN, S_BEFORE_KICK_OFF, S_KICK_OFF, S_KICK_IN,
	       S_CORNER_KICK, S_OFFSIDE, S_GOAL_KICK, S_FREE_KICK, 
	       S_PLAY_ON, S_HALF_TIME, S_TIME_UP, S_EXTEND, S_FOUL,
	       S_GOAL } rc_state_t;
typedef enum { VQ_HIGH, VQ_LOW } rc_view_quality_t;
typedef enum { VW_NARROW, VW_NORMAL, VW_WIDE } rc_view_width_t;

rc_result_t analyze(char *p);

#endif /* __PARSER_H__ */

/* End of file included parser.h */
/* File parserdefs.h included by the generator c2ecl */

#ifndef __PARSERDEFS_H__
#define __PARSERDEFS_H__

#define ON_SEE(TIME) step_num = TIME;
#define	ON_SEE_PLAYER(TEAMNAME, NUM, DIST, DIR, DISTCHG, DIRCHG, BODYDIR, HEADDIR) ;
#define ON_SEE_GOAL(SIDE, DIST, DIR) ;
#define ON_SEE_LINE(SIDE, DIST, DIR) ;
#define ON_SEE_FLAG(T1, T2, T3, DIST, DIR, DISTCHG, DIRCHG) ;
#define ON_SEE_BALL(DIST, DIR, DISTCHG, DIRCHG) ;

#define ON_HEAR(TIME, DIR, MSG) step_num = TIME;
#define ON_REFEREE(TIME, STATE, SIDE, NUM) step_num = TIME;

#define ON_SENSE_BODY(TIME, QUALITY, VIEWWIDTH, STAMINA, EFFORT, SPEED, HEADDIR, KICKCOUNT, DASHCOUNT, TURNCOUNT, SAYCOUNT, TURNNECKCOUNT) step_num = TIME; \\
percept_sense_body.time = TIME; \\
percept_sense_body.quality = QUALITY; \\
percept_sense_body.viewwidth = VIEWWIDTH; \\
percept_sense_body.stamina = STAMINA; \\
percept_sense_body.effort = EFFORT; \\
percept_sense_body.speed = SPEED; \\
percept_sense_body.headdir = HEADDIR; \\
percept_sense_body.kickcount = KICKCOUNT; \\
percept_sense_body.dashcount = DASHCOUNT; \\
percept_sense_body.turncount = TURNCOUNT; \\
percept_sense_body.saycount = SAYCOUNT; \\
percept_sense_body.turnneckcount = TURNNECKCOUNT;

#endif /* __PARSERDEFS_H__ */


/* End of file included parserdefs.h */
")

;;; Define and static variables
(clines "
#define MAX_CMD_LENGTH 100
#define QUEUE_SZ 6
#define ALPHA_TRASH 0.7
#define BETA_TRASH 0.3
#define WARN_THRESOLD 0.3
#define INTERVAL 150000
#define MAX_WAIT_PERCEPT 5
#define BOOST_CHECK 10
#define SUCC(x) ((x + 1) % QUEUE_SZ)

typedef void (*sighandler_t)(int);

static struct socket {
    int			socketfd;
    struct sockaddr_in	serv_addr;
} _sock, *sock = &_sock;

static struct rc_queue {
  int head;
  int tail;
  char buffer[QUEUE_SZ][MAX_CMD_LENGTH];
} _q, *q = &_q;

static struct __sense_body {
  int time;
  rc_view_quality_t quality;
  rc_view_width_t viewwidth;
  int stamina;
  int effort;
  float speed;
  float headdir;
  int kickcount;
  int dashcount;
  int turncount;
  int saycount;
  int turnneckcount;
} percept_sense_body;

static struct itimerval step = { { 0, INTERVAL }, { 0, INTERVAL }};
static float trashing = 0;
static int verbose = 0;
static int player_num = 0;
static int step_num = 0;
static int turn_trash = 1;
static int no_sense_body = 1;
static int boosting = 1;
static int fired_command = 0;
static int check_boost = BOOST_CHECK;

/* Statistics */
static int kick_count = 0;
static int turn_count = 0;
static int dash_count = 0;
static int say_count  = 0;
static int turn_neck_count = 0;

static int all_kick_count = 0;
static int all_turn_count = 0;
static int all_dash_count = 0;
static int all_say_count  = 0;
static int all_turn_neck_count = 0;
")

;;; Parser
(clines "
/*******************************************

 *******************************************/

#include <ctype.h>
#include <stdio.h>

rc_result_t analyze(char *p) {
  char *input = p;
  rc_obj_t objname;
  int i, time, num;
  char teamname[MAX_TEAMNAME_LEN], labels[3];
  float dist, dir, distchg, dirchg, bodydir, headdir, speed;
  rc_state_t state;
  char msg[MAX_MSG_LEN];
  rc_view_quality_t viewQuality;
  rc_view_width_t viewWidth;
  int stamina, effort, kickCount, dashCount, turnCount, 
      sayCount, turnNeckCount;

  switch (input[3]) {
  case 'e': /* see */
    input += 5;

    READ_INT(time);

    ON_SEE(time);

    while (*input != ')') { /* Read the object info */
      labels[0] = labels[1] = labels[2] = '\\0';
      dist = dir = distchg = dirchg = bodydir = headdir = INVALID_VALUE;

      input += 3; /* eat \" ((\" */

      switch (*input) {
      case 'p': /* player */
	objname = OBJ_PLAYER;
	input += 6;

	num = INVALID_VALUE;
	teamname[0] = '\\0';

	if (*input++ == ')')
	  break;

	i = 0;
	while (!isspace(*input) && (*input != ')')) ++i, ++input;

	strncpy(teamname, input - i, i);

	teamname[i] = '\\0';

	if (*input++ == ')')
	  break;

	READ_INT(num);
	break;
      case 'g': /* goal */
	objname = OBJ_GOAL;
	labels[0] = input[5];

	input += 7;
	break;
      case 'b': /* ball */
	objname = OBJ_BALL;
	input += 5;
	break;
      case 'f': /* flag */
	objname = OBJ_FLAG;
	input += 5;

	labels[0] = *input++;

	if (*input++ == ')')
	  break;

	labels[1] = *input++;

	if (*input++ == ')')
	  break;

	labels[2] = *input++;

	if (*input == ')')
	  ++input;
	else
	  input += 2;
	break;
      case 'l': /* line */
	objname = OBJ_LINE;
	labels[0] = input[5];

	input += 7;
	break;
      }

      if (*input != ')') {
	++input;
      
	READ_NUM(dist);

	if (*input != ')') {
	  ++input;

	  READ_NUM(dir);
	  
	  if (*input != ')') {
	    ++input;
	    
	    READ_NUM(distchg);
	    
	    if (*input != ')') {
	      ++input;

	      READ_NUM(dirchg);
	      
	      if (*input != ')') {
		++input;

		READ_NUM(bodydir);

		if (*input != ')') {
		  ++input;

		  READ_NUM(headdir);
		}
	      }
	    }
	  }
	}
      }

      ++input;

      switch (objname) {
      case OBJ_PLAYER:
	ON_SEE_PLAYER(teamname, num, dist, dir, distchg, dirchg, bodydir, headdir);
	break;
      case OBJ_GOAL:
	ON_SEE_GOAL(labels[0], dist, dir);
	break;
      case OBJ_BALL:
	ON_SEE_BALL(dist, dir, distchg, dirchg);
	break;
      case OBJ_FLAG:
	ON_SEE_FLAG(labels[0], labels[1], labels[2], dist, dir, distchg, dirchg);
	break;
      case OBJ_LINE:
	ON_SEE_LINE(labels[0], dist, dir);
	break;
      }
    }
    break;
  case 'a': /* hear */
    input += 6;

    READ_INT(time);

    ++input; /* white space */

    switch (*input) {
    case 's':
      dir = SELF_DIR;
      input += 5;
      break;
    case 'r':
      dir = REFEREE_DIR;
      input += 8;
      break;
    default:
      READ_NUM(dir);
      ++input;
    }

    if (dir == REFEREE_DIR) {
      labels[0] = '\\0';
      num = -1;

      switch (*input) {
      case 'b': /* before_kick_off */
	state = S_BEFORE_KICK_OFF;
	break;
      case 'k': /* kick_* */
	if (input[5] == 'i') {
	  state = S_KICK_IN;
	  labels[0] = input[8];
	} else {
	  state = S_KICK_OFF;
	  labels[0] = input[9];
	}
	break;
      case 'c': /* corner_kick */
	state = S_CORNER_KICK;
	labels[0] = input[12];
	break;
      case 'o': /* offside */
	state = S_OFFSIDE;
	labels[0] = input[8];
	break;
      case 'g': /* goal_* */
	if (input[5] == 'k') {
	  state = S_GOAL_KICK;
	  labels[0] = input[10];
	} else {
	  state = S_GOAL;
	  labels[0] = input[5];
	  input += 7;
	  READ_INT(num);
	}
	break;
      case 'f': /* f* */
	if (input[1] == 'r') {
	  state = S_FREE_KICK;
	  labels[0] = input[10];
	} else {
	  state = S_FOUL;
	  labels[0] = input[5];
	  input += 7;
	  READ_INT(num);
	}
	break;
      case 'p': /* play_on */
	state = S_PLAY_ON;
	break;
      case 'h': /* half_time */
	state = S_HALF_TIME;
	break;
      case 't': /* time_up */
	state = S_TIME_UP;
	break;
      case 'e': /* extend */
	state = S_EXTEND;
	break;
      default:
	state = S_UNKNOWN;
      }

      ON_REFEREE(time, state, labels[0], num);
    } else {
      strcpy(msg, input);
      msg[strlen(msg) - 1] = '\\0';

      ON_HEAR(time, dir, msg);
    }
    break;
  case 'n': /* sense_body */
    input += 12;

    READ_INT(time);

    ++input; /* white space */

    /* view_mode */
    input += 11;

    switch (*input) {
    case 'h':
      viewQuality = VQ_HIGH;
      input += 5;
      break;
    case 'l':
      viewQuality = VQ_LOW;
      input += 4;
      break;
    }

    switch (input[1]) {
    case 'a':
      viewWidth = VW_NARROW;
      input += 7;
      break;
    case 'o':
      viewWidth = VW_NORMAL;
      input += 7;
      break;
    case 'i':
      viewWidth = VW_WIDE;
      input += 5;
      break;
    }

    /* stamina */
    input += 10;

    READ_INT(stamina);

    ++input;

    READ_INT(effort);

    ++input;

    /* speed */
    input += 8;

    READ_NUM(speed);

    ++input;

    /* head_angle */
    input += 13;
    
    READ_NUM(headdir);

    ++input;

    /* kick */
    input += 7;

    READ_INT(kickCount);

    ++input;

    /* dash */
    input += 7;

    READ_INT(dashCount);

    ++input;

    /* turn */
    input += 7;

    READ_INT(turnCount);

    ++input;

    /* say */
    input += 6;

    READ_INT(sayCount);

    ++input;

    /* turn_neck */
    input += 12;

    READ_INT(turnNeckCount);

    ++input;

    ON_SENSE_BODY(time, viewQuality, viewWidth, stamina, effort, speed, headdir, kickCount, dashCount, turnCount, sayCount, turnNeckCount);
    break;
  }

  return OK;
}


")

;;; Header del file C generato dal compilatore
(clines "
int readNumber(char* percept) {
  char *p = percept;
  int ret = 0;

  while (*p && *p++ != ' ');

  while (isdigit(*p))
    ret = (ret * 10) + (*p++ - '0');

  return ret;
}

void add_cmd(const char *cmd) {
  int tt = 0;
  int tail = SUCC(q->tail);

  /* in case of trashing trash the oldest action */
  if (tail == q->head) {
/*    q->head = SUCC(q->head); */
    q->tail = q->head = 0;
    tail = 1;
  }

  if (!strncmp(cmd + 1, \"turn\", 4)) {
    char *c = cmd + 6;

    if (*c == '-') ++c;

    while (isdigit(*c))
      tt = (tt * 10) + (*c++ - '0');
  }

  /* If the command is a turn command reset the buffer */
  if (turn_trash && (tt > 10)) {
    q->tail = q->head = 0;
    tail = 1;
  }

  strcpy(q->buffer[q->tail], cmd);
  q->tail = tail;
}

char* get_cmd() {
  int head = q->head;

  if (q->tail == head) return NULL;

  q->head = SUCC(q->head);
  return q->buffer[head];
}

void fire(int type) {
  char *cmd;
  int n;
  int onecmd;

  while (setitimer(ITIMER_REAL, &step, NULL) == -1);
  if (signal(SIGALRM, (sighandler_t)fire) == SIG_ERR) {
    fprintf(stderr, \"Error installing the signal!\");
    exit(1);
  }

  onecmd = SUCC(q->head) == q->tail;

  cmd = get_cmd();

  if (!cmd) return;

  switch (cmd[1]) {
  case 'k':
    ++kick_count;
    ++all_kick_count;
    if (onecmd)
      fired_command = 1;
    break;
  case 'd':
    ++dash_count;
    ++all_dash_count;
    if (onecmd)
      fired_command = 1;
    break;
  case 't': // AGGIUSTA
    ++turn_count;
    ++all_turn_count;
    if (onecmd)
      fired_command = 1;
    break;
  case 's':
    ++say_count;
    ++all_say_count;
    if (onecmd)
      fired_command = 1;
    break;
  }

  n = strlen(cmd);

  if (sendto(sock->socketfd, cmd, n, 0,
	        (struct sockaddr *)&sock->serv_addr, sizeof(sock->serv_addr))
	  != n) {
    fprintf(stderr, \"Error sending data!\");
    exit(1);
  }
}

void adjustTimer(char* sense) {
  static int temp, new_val;

  analyze(sense);

  if (!fired_command) return;

  fired_command = 0;
  temp = 0;

//  if (player_num == 2)
//    fprintf(stderr, \"Sense body is %s\nBoosting is at %i, temp is %i. Dash count is (%i, %i)\\n\", sense, step.it_interval.tv_usec, temp, all_dash_count, percept_sense_body.dashcount);

  temp += percept_sense_body.kickcount - kick_count;
  if (temp != -1)
    kick_count = percept_sense_body.kickcount;

  temp += percept_sense_body.dashcount -  dash_count;
  if (temp != -1)
    dash_count = percept_sense_body.dashcount;

  temp += percept_sense_body.turnneckcount -  turn_neck_count;
  if (temp != -1)
    turn_neck_count = percept_sense_body.turnneckcount;

  temp += percept_sense_body.turncount -  turn_count;
  if (temp != -1)
    turn_count = percept_sense_body.turncount;

  temp += percept_sense_body.saycount -  say_count;
  if (temp != -1)
    say_count = percept_sense_body.saycount;          

  if (temp != -1) {
    if (boosting) {
      if (temp < -1) {
        boosting = 0;
        check_boost = BOOST_CHECK;
        step.it_interval.tv_usec += 1000;
        step.it_value.tv_usec += 1000;
      } else {
        step.it_interval.tv_usec -= 2000;
        step.it_value.tv_usec -= 2000;
      }
    } else {
      if (temp < -1) {
        step.it_interval.tv_usec += 1000;
        step.it_value.tv_usec += 1000;
      } else {
        if (check_boost)
          --check_boost;
        else
          boosting = 1;
      }
    }
  }
}
")

;;; Abilita la segnalazione su stdout del trashing
;;; @return nil
(defcbody rc-setitimer-interval (fixnum fixnum) object
"
  Cnil;
  {
    step.it_interval.tv_sec = #0;
    step.it_interval.tv_usec = #1;
    step.it_value.tv_sec = #0;
    step.it_value.tv_usec = #1;

    VALUES(0) = Cnil;

    RETURN(1);
  }
"
)

;;; Si connette al server
;;; @param object La stringa contenente il nome del server
;;; @param object La stringa contenente la porta a cui connettersi
(defcbody connect-server-soccer (object object) object
"
  Cnil;
  {
    struct hostent 	*host_ent;
    struct in_addr      *addr_ptr;
    struct sockaddr_in	cli_addr;
    int			sockfd, val;

    int serverPort = atoi(#1->st.st_self);

    if (signal(SIGALRM, (sighandler_t)fire) == SIG_ERR) {
      fprintf(stderr, \"Error installing the signal!\");
      exit(1);
    }

    if  (setitimer(ITIMER_REAL, &step, NULL) == -1) {
	fprintf(stderr, \"Error installing the setitimer handler!\\n\");
        exit(1);
    }

    q->tail = q->head = 0;
    sock->socketfd = -1;

    if ((host_ent = (struct hostent *)gethostbyname(#0->st.st_self)) == NULL) {
	/* Check if a numeric address */
	if (inet_addr(#0->st.st_self) == -1) {
	    VALUES(0) = Cnil;
        }
    } else {
	addr_ptr = (struct in_addr *) *host_ent->h_addr_list;
	#0->st.st_self = inet_ntoa(*addr_ptr);
    }


    /*
     *  Open UDP socket.
     */
    if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	VALUES(0) = Cnil;  	/* Can't open socket. */
    }

    val = fcntl(sockfd, F_GETFL, 0);
    val |= O_NONBLOCK;
    fcntl(sockfd, F_SETFL, val);

    /*
     *  Bind any local address.
     */
    memset((char *) &cli_addr, 0, sizeof(cli_addr));
    cli_addr.sin_family		= AF_INET;
    cli_addr.sin_addr.s_addr	= htonl(INADDR_ANY);
    cli_addr.sin_port		= htons(0);

    if (bind(sockfd, (struct sockaddr *) &cli_addr,
	    sizeof(cli_addr)) < 0) {
	VALUES(0) = Cnil;  /* Can't bind local address */
    }

    /*
     *  Fill in the structure with the address of the server.
     */
    sock->socketfd = sockfd;

    memset((char *) &sock->serv_addr, 0, sizeof(sock->serv_addr));
    sock->serv_addr.sin_family		= AF_INET;
    sock->serv_addr.sin_addr.s_addr	= inet_addr(#0->st.st_self);
    sock->serv_addr.sin_port		= htons(serverPort);

    VALUES(0) = Ct;

    RETURN(1);
  } 
"
)

;;; Invia un messaggio al server
;;; @param object Stringa da inviare al server
;;; @return T se il pacchetto e' stato inviato nil altrimenti.
(defcbody send-msg (object) object
"
  Cnil;
  {
    add_cmd(#0->st.st_self);

    VALUES(0) = Ct;

    RETURN(1);
  }
"
)

;;; Riceve un pacchetto dal server
;;; @return La stringa contenente il messaggio
(defcbody receive-msg () object
"
  Cnil;
  {
    fd_set              desc;
    struct timeval      tv;
    int			n, servlen, retval;
    struct sockaddr_in	serv_addr;
    char buf[2048];

    object ret;
    
    servlen = sizeof(serv_addr);

    /* Wait for data... */
    FD_ZERO(&desc);
    FD_SET(sock->socketfd, &desc);

    tv.tv_sec = MAX_WAIT_PERCEPT;
    tv.tv_usec = 0;

    retval = select(1, &desc, NULL, NULL, &tv);

    if (!retval) {
        printf(\"Server seems to be down, exiting...\\n\");
        exit(0);
    }

    do {
      n = recvfrom(sock->socketfd, buf, 2048, 0,
			 (struct sockaddr *)&serv_addr, &servlen);

      buf[n] = '\\0';

      if (buf[6] == '_')
        adjustTimer(buf);

      if ((n != -1) && no_sense_body && !strncmp(buf + 1, \"sense\", 5))
	n = -1;
    } while (n == -1); /* blocca */

    sock->serv_addr.sin_port = serv_addr.sin_port;
   
    analyze(buf);

    ret = (object)make_string_copy(buf);
    VALUES(0) = ret;

    RETURN(1);
  }
"
)

;;; Riceve l'ultimo messaggio inviato dal server ignorando tutti i precedenti
;;; @return La stringa contenente il messaggio
(defcbody receive-last-msg () object
"
  Cnil;
  {
    fd_set              desc;
    struct timeval      tv;
    int			n, servlen, trashed = -1, retval;
    struct sockaddr_in	serv_addr;
    char buf[2048], retstr[2048];

    object ret;

    servlen = sizeof(serv_addr);

    retstr[0] = '\\0';

    /* Wait for data... */
    FD_ZERO(&desc);
    FD_SET(sock->socketfd, &desc);

    tv.tv_sec = MAX_WAIT_PERCEPT;
    tv.tv_usec = 0;

    retval = select(1, &desc, NULL, NULL, &tv);

    if (!retval) {
        printf(\"Server seems to be down, exiting...\\n\");
        exit(0);
    }

    do {
      n = recvfrom(sock->socketfd, buf, 2048, 0,
			 (struct sockaddr *)&serv_addr, &servlen);

      if (n != -1) {
        ++trashed;
        buf[n] = '\\0';
        strcpy(retstr, buf);
        if (!strncmp(retstr + 1, \"hear\", 4)) {
          errno = EWOULDBLOCK;
          break;
        } 
      }

      if (buf[6] == '_')
        adjustTimer(buf);

      if ((n != -1) && no_sense_body && !strncmp(buf + 1, \"sense\", 5)) {
        n = -1;
        retstr[0] = '\\0';
      }
    } while(!retstr[0] || (n != -1)); /* blocca */

    trashing = (float)((ALPHA_TRASH * trashed) + (BETA_TRASH * trashing));
    if (verbose && (trashing > WARN_THRESOLD))
      printf(\"Player number %i has a trashing of %f at cycle step %i\\\n\", player_num, trashing, step_num);

    if (errno != EWOULDBLOCK) {
	VALUES(0) = Cnil;
    } else {
      sock->serv_addr.sin_port = serv_addr.sin_port;
      retstr[n] = '\\0';

      analyze(retstr);

      ret = (object)make_string_copy(retstr);
      VALUES(0) = ret;
    }

    RETURN(1);
  }
"
)

;;; Chiude la connessione col server
;;; @return nil
(defcbody close-connection () object
"
  Cnil;
  {
    close(sock->socketfd) ;

    VALUES(0) = Cnil;

    RETURN(1);
  }
"
)

;;; Restituisce la statistica corrente sul numero di pacchetti
;;; scartati.
;;; @return Il numero di pacchetti scartati
(defcbody rc-trashing () object
"
  Cnil;
  {
    VALUES(0) = make_shortfloat(trashing);

    RETURN(1);
  }
"
)

;;; Abilita la segnalazione su stdout del trashing
;;; @return nil
(defcbody rc-verbose-on () object
"
  Cnil;
  {
    verbose = 1;
    VALUES(0) = Cnil;

    RETURN(1);
  }
"
)

;;; Disabilita la segnalazione su stdout del trashing
;;; @return nil
(defcbody rc-verbose-off () object
"
  Cnil;
  {
    verbose = 0;
    VALUES(0) = Cnil;

    RETURN(1);
  }
"
)

;;; Disabilita il trash della coda all'arrivo di un turn.
;;; Necessario per le sequenze di azioni coordinate
;;; @return nil
(defcbody rc-turn-trash-off () object
"
  Cnil;
  {
    turn_trash = 0;
    VALUES(0) = Cnil;

    RETURN(1);
  }
"
)

;;; Disabilita il trash della coda all'arrivo di un turn.
;;; Necessario per le sequenze di azioni coordinate
;;; @return nil
(defcbody rc-turn-trash-on () object
"
  Cnil;
  {
    turn_trash = 1;
    VALUES(0) = Cnil;

    RETURN(1);
  }
"
)

;;; Abilita la ricezione di messaggi sense-body
;;; @return nil
(defcbody rc-sense-body-on () object
"
  Cnil;
  {
    no_sense_body = 0;
    VALUES(0) = Cnil;

    RETURN(1);
  }
"
)

;;; Disabilita la ricezione di messaggi sense-body
;;; @return nil
(defcbody rc-sense-body-off () object
"
  Cnil;
  {
    no_sense_body = 1;
    VALUES(0) = Cnil;

    RETURN(1);
  }
"
)

;;; Imposta il numero del giocatore nel codice C
;;; Viene utilizzato solo per scopi di warning
(defcbody rc-set-jersey-num (fixnum) object
"
  Cnil;
  {
    player_num = #0;
    VALUES(0) = Cnil;
    RETURN(1);
  }
"
)

(defun generic-loader (base file)
  "Generic code loader given a specific codebase."
  (let ((fn (format nil "~A/~A" base file)))
    ;; Try to load the file
    (unless (load fn :if-does-not-exist nil) ; Can't find!
      (unless (load (format nil "~A.o" fn) :if-does-not-exist nil) ; no .o
	(load (format nil "~A.lsp" fn))))
    t))

(defun rcload (lib)
  (generic-loader *robocup-lib* lib))

;;; Queste funzioni elevano il livello della connessione con il lisp

(defun connect-to-server (host &optional (port 6000))
  "Si collega al server
  @param host Stringa contenente il nome della macchina su cui gira il server
  @param port Porta a cui connettersi. Se omessa viene assunto il valore 6000
  @return T se la connessione e' avvenuta correttamente NIL altrimenti"
  (if (and (< port 65536) (> port 0))
      (if (connect-server-soccer host (princ-to-string port))
	  T
	NIL)
    NIL))

(defun read-percept ()
  "Legge un messaggio dal server e lo converte in una S-Expr
  @return L'S-Expr o NIL se qualcosa e' andato male"
  (let ((msg (receive-msg)))
    (if msg
	(read-from-string msg)
      NIL)))

(defun read-last-percept()
  "legge l'ultimo messaggio inviato dal server, ignorando i precedenti,
   e restituisce l's-expr
   @return L'S-Expr o NIL se qualcosa e' andato male"
  (let ((msg (receive-last-msg)))
    (if msg
	(read-from-string msg)
      NIL)))

(defun do-action (action)
"Invia una s-expr al server
  @param action S-Expr da inviare al server
  @return T se l'S-Expr e' stato inviato nil altrimenti."
  (send-msg (string-downcase (princ-to-string action))))

(defun do-actions (action-list)
"Invia una lista di comandi al server
  @param action S-Expr da inviare al server
  @return T se l'S-Expr e' stato inviato nil altrimenti."
(dolist (action action-list)
  (send-msg (string-downcase (princ-to-string action)))))

(defun end-connection ()
"Chiude la connessione col server
  @return NIL"
  (close-connection))
