;;; Where libraries are stored
(defconstant *robocup-lib* "/home/cisterni/RoboCup/teams/IALP/lib")

(clines "
#include <ctype.h>
#include <stdio.h>
#include \"parser.h\"
#include \"parserdefs.h\"
#include \"parser.h\"
#include \"queue.h\"
#include \"smartlow.h\"

/** File parser.c begins here */
/****************************************************************

  Author: Antonio Cisternino

  Version: 1.5.24

  Description: This file contains the parser code. See the 
               parserdefs.h file to hook the parser events.

 ****************************************************************/






rc_result_t analyze(char *p) {
  char *input = p;
  rc_obj_t objname;
  int i, time, num;
  char teamname[MAX_TEAMNAME_LEN], labels[3];
  float dist, dir, distchg, dirchg, bodydir, headdir, speed;
  rc_state_t state;
  char msg[MAX_MSG_LEN];
  rc_view_quality_t viewQuality;
  rc_view_width_t viewWidth;
  rc_error_t error_type;
  int stamina, effort, kickCount, dashCount, turnCount, 
      sayCount, turnNeckCount;

  switch (input[3]) {
  case 'i': /* init */
    input += 6;

    labels[0] = *input;
    input += 2;

    READ_INT(num);
    ++input;

    CHECK_STATE(labels[0]);

    ON_INIT(labels[0], num, state);
    break;
  case 'r': /* error */
    input += 7;

    if (*input == 'u') { /* unknown_command */
      error_type = E_UNKNOWN_COMMAND;
    } else { /* illigal_command_form */
      error_type = E_ILLIGAL_COMMAND_FORM;
    }

    ON_ERROR(error_type);
    break;
  case 'e': /* see */
    input += 5;

    READ_INT(time);

    ON_SEE(time);

    while (*input != ')') { /* Read the object info */
      labels[0] = labels[1] = labels[2] = '\\0';
      dist = dir = distchg = dirchg = bodydir = headdir = INVALID_VALUE;

      input += 3; /* eat \" ((\" */

      switch (*input) {
      case 'p': /* player */
	objname = OBJ_PLAYER;
	input += 6;

	num = INVALID_VALUE;
	teamname[0] = '\\0';

	if (*input++ == ')')
	  break;

	i = 0;
	while (!isspace(*input) && (*input != ')')) ++i, ++input;

	strncpy(teamname, input - i, i);

	teamname[i] = '\\0';

	if (*input++ == ')')
	  break;

	READ_INT(num);
	break;
      case 'g': /* goal */
	objname = OBJ_GOAL;
	labels[0] = input[5];

	input += 7;
	break;
      case 'b': /* ball */
	objname = OBJ_BALL;
	input += 5;
	break;
      case 'f': /* flag */
	objname = OBJ_FLAG;
	input += 5;

	labels[0] = *input++;

	if (*input++ == ')')
	  break;

	labels[1] = *input++;

	if (*input++ == ')')
	  break;

	labels[2] = *input++;

	if (*input == ')')
	  ++input;
	else
	  input += 2;
	break;
      case 'l': /* line */
	objname = OBJ_LINE;
	labels[0] = input[5];

	input += 7;
	break;
      }

      if (*input != ')') {
	++input;
      
	READ_NUM(dist);

	if (*input != ')') {
	  ++input;

	  READ_NUM(dir);
	  
	  if (*input != ')') {
	    ++input;
	    
	    READ_NUM(distchg);
	    
	    if (*input != ')') {
	      ++input;

	      READ_NUM(dirchg);
	      
	      if (*input != ')') {
		++input;

		READ_NUM(bodydir);

		if (*input != ')') {
		  ++input;

		  READ_NUM(headdir);
		}
	      }
	    }
	  }
	}
      }

      ++input;

      switch (objname) {
      case OBJ_PLAYER:
	ON_SEE_PLAYER(teamname, num, dist, dir, distchg, dirchg, bodydir, headdir);
	break;
      case OBJ_GOAL:
	ON_SEE_GOAL(labels[0], dist, dir);
	break;
      case OBJ_BALL:
	ON_SEE_BALL(dist, dir, distchg, dirchg);
	break;
      case OBJ_FLAG:
	ON_SEE_FLAG(labels[0], labels[1], labels[2], dist, dir, distchg, dirchg);
	break;
      case OBJ_LINE:
	ON_SEE_LINE(labels[0], dist, dir);
	break;
      }
    }
    break;
  case 'a': /* hear */
    input += 6;

    READ_INT(time);

    ++input; /* white space */

    switch (*input) {
    case 's':
      dir = SELF_DIR;
      input += 5;
      break;
    case 'r':
      dir = REFEREE_DIR;
      input += 8;
      break;
    default:
      READ_NUM(dir);
      ++input;
    }

    if (dir == REFEREE_DIR) {
      labels[0] = '\\0';
      num = -1;

      CHECK_STATE(labels[0]);

      ON_REFEREE(time, state, labels[0], num);
    } else {
      strcpy(msg, input);
      msg[strlen(msg) - 1] = '\\0';

      ON_HEAR(time, dir, msg);
    }
    break;
  case 'n': /* sense_body */
    input += 12;

    READ_INT(time);

    ++input; /* white space */

    /* view_mode */
    input += 11;

    switch (*input) {
    case 'h':
      viewQuality = VQ_HIGH;
      input += 5;
      break;
    case 'l':
      viewQuality = VQ_LOW;
      input += 4;
      break;
    }

    switch (input[1]) {
    case 'a':
      viewWidth = VW_NARROW;
      input += 7;
      break;
    case 'o':
      viewWidth = VW_NORMAL;
      input += 7;
      break;
    case 'i':
      viewWidth = VW_WIDE;
      input += 5;
      break;
    }

    /* stamina */
    input += 10;

    READ_INT(stamina);

    ++input;

    READ_INT(effort);

    ++input;

    /* speed */
    input += 8;

    READ_NUM(speed);

    ++input;

    /* head_angle */
    input += 13;
    
    READ_NUM(headdir);

    ++input;

    /* kick */
    input += 7;

    READ_INT(kickCount);

    ++input;

    /* dash */
    input += 7;

    READ_INT(dashCount);

    ++input;

    /* turn */
    input += 7;

    READ_INT(turnCount);

    ++input;

    /* say */
    input += 6;

    READ_INT(sayCount);

    ++input;

    /* turn_neck */
    input += 12;

    READ_INT(turnNeckCount);

    ++input;

    ON_SENSE_BODY(time, viewQuality, viewWidth, stamina, effort, speed, headdir, kickCount, dashCount, turnCount, sayCount, turnNeckCount);
    break;
  }

  return OK;
}


/** File parser.c ends here */

/** File queue.c begins here */



/** Commands' queue */
static smart_queue  _q, *q = &_q;

/**
 * Return the length of the queue.
 * @return The length of the queue.
 */
int smart_queue_length() {
  return (q->tail < q->head)?(q->tail + QUEUE_SZ - q->head):
    (q->tail - q->head); 
}

/**
 * Discard all the commands in the queue and set the queue's length
 * to zero.
 */
void smart_reset_queue() {
  q->tail = q->head = 0;
}

/**
 * Add a command to the queue. The command is represented by a
 * common string. In this function is implemented the turn
 * trashing.
 * @param cmd Command that must be enqueued
 * @see smart_turn_trashing(int)
 */
void smart_add_queue(const char *cmd) {
  int tt = 0, tail = SUCC(q->tail);

  /* If the queue is full a reset is done */
  if (tail == q->head) {
    q->tail = q->head = 0;
    tail = 1;
  }

  /* Check for turn command */
  if (!strncmp(cmd + 1, \"turn\", 4)) {
    const char *input = cmd + 6;

    if (*input == '-') ++input;

    READ_INT(tt);
  }

  /* If the command is a turn command reset the buffer */
  if (turn_trash && (tt > 10)) {
    q->tail = q->head = 0;
    tail = 1;
  }

  /* Add the command to the queue */
  strcpy(q->buffer[q->tail], cmd);
  q->tail = tail;
}

/**
 * Get a command from the queue.
 * @return the first command in the queue or NULL if the queue is empty.
 */
char* smart_get_queue() {
  int head = q->head;

  if (q->tail == head) return NULL;

  q->head = SUCC(q->head);
  return q->buffer[head];
}

/** File queue.c ends here */

/** File smartlow.c begins here */


/** Array that contain the label used for change_view */
static char* angle_width_v[] = { \"narrow\", \"normal\", \"wide\" };
static char* quality_v[] = { \"high\", \"low\" };

/** Handler for the timer */
void smart_fire(int type);

/** Function that adjusts the value of the timer */
void smart_adjust_timer();

/** Contain the informations about the last sense_body perception.  */
static smart_sense_body percept_sense_body;

/** Type of the percept */
static smart_perception_type perception_type;

/** Contain the socket for the connection with the server soccer.  */
static smart_socket _sock, *sock = &_sock;

/** Initial step of the timer */
static struct itimerval step = { { 0, INTERVAL }, { 0, INTERVAL }};

/** Used to adjust dinamically the timer value */
static struct timeval last_sense_body = { 0, 0 };

/** If 1 when a turn command is sent the command queue is cleaned */
static int turn_trash = 1;

/** Buffer used for receiving a message from server  */
static char buf[MSG_BUF_SIZE];

/** Buffer used to send commands */
static char msgbuf[MAX_CMD_LENGTH];

/** Alpha coeff to set the timer value */
static float alpha = ALPHA;

/**
 * Open the connection with the server soccer.
 * @param host Hostname or IP address of the server
 *             where the soccerserver is running.
 * @param port Port used for the connection (usually 6000)
 * @return 0 is returned if the connection is ok -1 otherwise.
 */
int smart_connect_server_soccer(const char* host, const int port) {
  /* This code is taken from the sampleclient by Noda */
  struct hostent *host_ent;
  struct in_addr *addr_ptr;
  struct sockaddr_in cli_addr;
  int sockfd, val;

  /* Install the signal handler */
  if (signal(SIGALRM, (sighandler_t)smart_fire) == SIG_ERR) {
    fprintf(stderr, \"Error installing SIGALRM handler.\\n\");
    exit(1);
  }

  /* Install the timer */
  if  (setitimer(ITIMER_REAL, &step, NULL) == -1) {
    fprintf(stderr, \"Error installing the setitimer handler!\\n\");
    exit(1);
  }

  /* Reset the commands' queue */
  smart_reset_queue();
  sock->socketfd = -1;

  if ((host_ent = (struct hostent *)gethostbyname(host)) == NULL) {
    /* Check if a numeric address */
    if (inet_addr(host) == -1) {
      return -1; /* It is not a valid IP address nor a DNS name */
    }
  } else {
    addr_ptr = (struct in_addr *) *host_ent->h_addr_list;
    host = inet_ntoa(*addr_ptr);
  }

  /* Open UDP socket. */
  if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    /* Can't open socket. */
    return -1;
  }

  /* Make the socket non blocking */
  val = fcntl(sockfd, F_GETFL, 0);
  val |= O_NONBLOCK;
  fcntl(sockfd, F_SETFL, val);

  /* Bind any local address. */
  memset((char *) &cli_addr, 0, sizeof(cli_addr));
  cli_addr.sin_family = AF_INET;
  cli_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  cli_addr.sin_port = htons(0);

  if (bind(sockfd, (struct sockaddr *) &cli_addr,
	   sizeof(cli_addr)) < 0) {
    return -1;  /* Can't bind local address */
  }

  /* Fill in the structure with the address of the server. */
  sock->socketfd = sockfd;

  memset((char *) &sock->serv_addr, 0, sizeof(sock->serv_addr));
  sock->serv_addr.sin_family = AF_INET;
  sock->serv_addr.sin_addr.s_addr = inet_addr(host);
  sock->serv_addr.sin_port = htons(port);

  return 0;
} 

/**
 * Close the connection with server.
 */
void smart_close_connection() {
  close(sock->socketfd) ;
}

/**
 * Read a message from the server. This is a synchronous
 * call. She waits for the message.
 * @return A pointer to the buffer that contains the message.
 */
char* smart_read_percept()  {
  fd_set desc;
  struct timeval tv;
  int n, servlen, retval;
  struct sockaddr_in serv_addr;

  servlen = sizeof(serv_addr);

  /* Prepare the descriptor for the select */
  FD_ZERO(&desc);
  FD_SET(sock->socketfd, &desc);

  /* Set the timeout */
  tv.tv_sec = MAX_WAIT_PERCEPT;
  tv.tv_usec = 0;

  /* Wait for data */
  do {
    retval = select(sock->socketfd + 1, &desc, NULL, NULL, &tv);
  } while(retval == -1);

  if (!retval) {
    printf(\"Server seems to be down, exiting...\\n\");
    exit(0);
  }

  /* Receive the data and parse the message received */
  n = recvfrom(sock->socketfd, buf, MSG_BUF_SIZE, 0,
	       (struct sockaddr *)&serv_addr, &servlen);

  buf[n] = '\\0';

  analyze(buf);

  /* If the message is a SENSE_BODY try to adjust the timer */
  if (perception_type == SMPT_SENSE_BODY)
    smart_adjust_timer();

  sock->serv_addr.sin_port = serv_addr.sin_port;

  return buf;
}

/**
 * Read the last message from the server. Other messages
 * are discarded. If a (hear) message is found the message
 * is returned and the function ends. In this case maybe that
 * other messages are leaved on the socket queue. The rationale
 * behind this semantics is that (hear) messages may contain
 * referee messages that represent the state of the game.
 * @return A pointer to the buffer that contains the message.
 */
char* smart_read_last_percept()   {
  fd_set desc;
  struct timeval tv;
  int n, servlen, retval;
  struct sockaddr_in serv_addr;

  servlen = sizeof(serv_addr);

  /* Prepare the descriptor for the select */
  FD_ZERO(&desc);
  FD_SET(sock->socketfd, &desc);

  tv.tv_sec = MAX_WAIT_PERCEPT;
  tv.tv_usec = 0;

  /* Wait for data */
  do {
    retval = select(sock->socketfd + 1, &desc, NULL, NULL, &tv);
  } while (retval == -1);

  if (!retval) {
    printf(\"Server seems to be down, exiting...\\n\");
    exit(0);
  }

  tv.tv_sec = 0;

  /* Receive all the data enqueued on the socket */
  while (select(sock->socketfd + 1, &desc, NULL, NULL, &tv)) {
    n = recvfrom(sock->socketfd, buf, MSG_BUF_SIZE, 0,
		 (struct sockaddr *)&serv_addr, &servlen);

    buf[n] = '\\0';

    analyze(buf);

    /* If the message is a SENSE_BODY try to adjust the timer */
    if (perception_type == SMPT_SENSE_BODY)
      smart_adjust_timer();

    sock->serv_addr.sin_port = serv_addr.sin_port;

    /* Break if the message is hear or referee */
    if ((perception_type == SMPT_HEAR) ||
	(perception_type == SMPT_REFEREE))
      break;
  }

  return buf;
}

/**
 * Handler installed for the SIGALRM signal. Each time that is
 * invoked by the kernel a message to the server is sent if there
 * are message in the commands' queue.
 * @param type Parameter set by the kernel
 */
void smart_fire(int type) {
  char *cmd;
  int n;

  /* Reinstall the timer and the signal's handler
     The setitimer call it is not strictly required but it is
     used because for every sense_body perceived the interval
     may changes. */
  while (setitimer(ITIMER_REAL, &step, NULL) == -1);
  if (signal(SIGALRM, (sighandler_t)smart_fire) == SIG_ERR) {
    fprintf(stderr, \"Error installing the signal!\");
    exit(1);
  }

  /* Get a command from the queue and, if the queue is not empty,
     sends it to the server */
  cmd = smart_get_queue();

  if (!cmd) return;

  n = strlen(cmd);

  if ((sendto(sock->socketfd, cmd, n, 0,
	      (struct sockaddr *)&sock->serv_addr, sizeof(sock->serv_addr)))
	  != n) {
    fprintf(stderr, \"Error sending data!\\n\");
    exit(1);
  }
}

/**
 * Measure the temporal distance between two successive sense_body.
 * The value newval calculated in this way affects the timer interval
 * that becomes oldval * alpha + (1 - alpha) * newval.
 */
void smart_adjust_timer() {
  static struct timeval tv;
  static struct timezone tz;

  gettimeofday(&tv, &tz);

  if (last_sense_body.tv_sec || last_sense_body.tv_usec) {
    int newval = ((int)(tv.tv_sec - last_sense_body.tv_sec)) * 1000000 + (int)
      (tv.tv_usec - last_sense_body.tv_usec);
    step.it_interval.tv_usec = 
      step.it_value.tv_usec = (int)(alpha * step.it_value.tv_usec + (1 - alpha) * newval);
  }

  last_sense_body = tv;
}

/**
 * Send a command to the server. The second parameter specifies
 * the mode of transmission: using the SEND_NOW constant the
 * message is delivered immediatly; otherwise the SEND_QUEUE
 * constant indicates that the message must be enqueued and
 * sent when the timer expires.
 * The preffered mode is the second one because the timer
 * guarantees that a command is sent using the optimal timing
 * without loss of commands. For this reason the macro defined
 * for sending commands use this mode. The only exception is
 * the macro for sending (turn-neck) command because this command
 * is received by the server also if coupled with other commands.
 * @param cmd Command that must be sent to the server.
 * @param mode If it is SEND_NOW the command is immediatly sent;
 *             otherwise if it is SEND_QUEUE the command is added
 *             to the command's queue.
 */
void smart_send_command(char* cmd, smart_send_mode mode) {
  if (mode == SEND_NOW) {
    int n = strlen(cmd);

    if ((sendto(sock->socketfd, cmd, n, 0,
		(struct sockaddr *)&sock->serv_addr, sizeof(sock->serv_addr)))
	!= n) {
      fprintf(stderr, \"Error sending data!\\n\");
      exit(1);
    }
  } else
    smart_add_queue(cmd);
}

/**
 * Change the value for the ALPHA parameter.
 * @param a New value for alpha
 * @return The old value for the parameter.
 * @see ALPHA
 */
int smart_set_alpha_timer(float a) {
  if ((a > 0) && (a < 1)) {
    int old = alpha;

    alpha = a;
    return old;
  }

  return -1;
}

/**
 * Change the value for the turn_trash flag. When turn trashing is on
 * the commands' queue becomes empty before the insertion of a 'turn'
 * command with a moment greater than 10. The rationale for this behavior
 * is that turning changes substantially the world perceived by the
 * player and thus, if the internal state delivers a turn action means
 * that the expectation is that the action is performed immediately.
 * The insertion of a dash command in queue does not change too much
 * the world perceived and then the queue is preserved.
 * The turn trashing can be disable because sometime maybe useful
 * sending series of commands (i.e. for turning with ball) that
 * contains turn actions interleaved with other actions.
 * @param tt It can be 1 or 0 to enable or disable the turn trashing.
 * @return The previous value of the flag.
 */
int smart_set_turn_trash(int tt) {
  int old = turn_trash;

  turn_trash = (tt)?1:0;

  return old;
}

/**
 * Returns 1 if turn trashing is enable 0 otherwise.
 * @return The value of turn trashing flag.
 * @see smart_set_turn_trash(int)
 */
int smart_get_turn_trash() {
  return turn_trash;
}

/**
 * Check if a message from the server is available without
 * parsing the new message. To get the message use smart_read_percept()
 * or smart_read_last_percept().
 * @return 1 if the message is available 0 otherwise.
 * @see smart_read_percept()
 * @see smart_read_last_percept()
 */
int smart_is_percept_available() {
  fd_set desc;
  struct timeval tv = { 0, 0 };
  int retval;

  /* Wait for data... */
  FD_ZERO(&desc);
  FD_SET(sock->socketfd, &desc);

  do {
    retval = select(sock->socketfd + 1, &desc, NULL, NULL, &tv);
  } while(retval == -1);

  return retval;
}


/** File smartlow.c ends here */
")

;;; Si connette al server
;;; @param object La stringa contenente il nome del server
;;; @param object La stringa contenente la porta a cui connettersi
(defcbody connect-server-soccer (object object) object
"
  Cnil;
  {
    if (!smart_connect_server_soccer(#0->st.st_self, atoi(#1->st.st_self)))
      VALUES(0) = Ct;
    else
      VALUES(0) = Cnil;

    RETURN(1);
  } 
"
)

;;; Invia un messaggio al server
;;; @param object Stringa da inviare al server
;;; @return T se il pacchetto e' stato inviato nil altrimenti.
(defcbody send-msg (object) object
"
  Cnil;
  {
    smart_send_command(#0->st.st_self, SEND_QUEUE);

    VALUES(0) = Ct;

    RETURN(1);
  }
"
)

;;; Riceve un pacchetto dal server
;;; @return La stringa contenente il messaggio
(defcbody receive-msg () object
"
  Cnil;
  {
    VALUES(0) = (object)make_string_copy(smart_read_percept());

    RETURN(1);
  }
"
)

;;; Riceve l'ultimo messaggio inviato dal server ignorando tutti i precedenti
;;; @return La stringa contenente il messaggio
(defcbody receive-last-msg () object
"
  Cnil;
  {
    VALUES(0) = (object)make_string_copy(smart_read_last_percept());

    RETURN(1);
  }
"
)

;;; Chiude la connessione col server
;;; @return nil
(defcbody close-connection () object
"
  Cnil;
  {
    smart_close_connection();

    VALUES(0) = Cnil;

    RETURN(1);
  }
"
)

;;; Disabilita il trash della coda all'arrivo di un turn.
;;; Necessario per le sequenze di azioni coordinate
;;; @return nil
(defcbody rc-turn-trash-off () object
"
  Cnil;
  {
    smart_set_turn_trash(0);

    VALUES(0) = Cnil;

    RETURN(1);
  }
"
)

;;; Disabilita il trash della coda all'arrivo di un turn.
;;; Necessario per le sequenze di azioni coordinate
;;; @return nil
(defcbody rc-turn-trash-on () object
"
  Cnil;
  {
    smart_set_turn_trash(1);

    VALUES(0) = Cnil;

    RETURN(1);
  }
"
)

(defun generic-loader (base file)
  "Generic code loader given a specific codebase."
  (let ((fn (format nil "~A/~A" base file)))
    ;; Try to load the file
    (unless (load fn :if-does-not-exist nil) ; Can't find!
      (unless (load (format nil "~A.o" fn) :if-does-not-exist nil) ; no .o
	(load (format nil "~A.lsp" fn))))
    t))

(defun rcload (lib)
  (generic-loader *robocup-lib* lib))

;;; Queste funzioni elevano il livello della connessione con il lisp

(defun connect-to-server (host &optional (port 6000))
  "Si collega al server
  @param host Stringa contenente il nome della macchina su cui gira il server
  @param port Porta a cui connettersi. Se omessa viene assunto il valore 6000
  @return T se la connessione e' avvenuta correttamente NIL altrimenti"
  (if (and (< port 65536) (> port 0))
      (if (connect-server-soccer host (princ-to-string port))
	  T
	NIL)
    NIL))

(defun read-percept ()
  "Legge un messaggio dal server e lo converte in una S-Expr
  @return L'S-Expr o NIL se qualcosa e' andato male"
  (let ((msg (receive-msg)))
    (if msg
	(read-from-string msg)
      NIL)))

(defun read-last-percept()
  "legge l'ultimo messaggio inviato dal server, ignorando i precedenti,
   e restituisce l's-expr
   @return L'S-Expr o NIL se qualcosa e' andato male"
  (let ((msg (receive-last-msg)))
    (if msg
	(read-from-string msg)
      NIL)))

(defun do-action (action)
"Invia una s-expr al server
  @param action S-Expr da inviare al server
  @return T se l'S-Expr e' stato inviato nil altrimenti."
  (send-msg (string-downcase (princ-to-string action))))

(defun do-actions (action-list)
"Invia una lista di comandi al server
  @param action S-Expr da inviare al server
  @return T se l'S-Expr e' stato inviato nil altrimenti."
(dolist (action action-list)
  (send-msg (string-downcase (princ-to-string action)))))

(defun end-connection ()
"Chiude la connessione col server
  @return NIL"
  (close-connection))

