;;;;
;;;; Last update: 2/1/1999
;;;; @author Antonio Cisternino
;;;; @version 0.2
;;;;
;;;; Functions that calculates positions of objects and of the agent.
;;;;
;;;; Changes:
;;;;
;;;; 14/3/1999: Fixed a serious bug: when the team play on right side
;;;;            left and right must be exchanged like top and bottom
;;;;
;;;; 2/1/1999: Added a lot of comments
;;;;
;;;; 2/1/1999: Fixed a bug of ECL on Linux

;;; Calculate the absolute player position
(defun calc-player-pos (kb p)
  "This method calculate the absolute position of the player using
   the fix points such as lines or line flags"


  ;; Check that the percept is of the type 'SEE
  (when (eq 'see (car p))

    (let (my-left my-right my-top my-bottom)
      (if (eq (agent-kb-side kb) #\L)
	  (setf
	   my-left 'l
	   my-right 'r
	   my-top 't
	   my-bottom 'b)
	(setf
	 my-left 'r
	 my-right 'l
	 my-top 'b
	 my-bottom 't))
      
      ;; x, y and theta are the result of the calculus
      (let ((x nil) (y nil) (theta nil) gamma)

	;; first of all we look for a line to calculate the direction
	(dolist 
	 (el (cddr p))

	 ;; (LINE ...)
	 (when (eq 'line (caar el))
	   (cond

	    ;; Left line: I can calculate the x and theta
	    ((eq my-left (cadar el))
	     (setf x (+ (- 0 half-pitch-length) 
			(abs (* (second el) (deg-sin (third el))))))
	     (setf theta (if (> (third el) 0)
			     (- -90 (third el))
			   (- 90 (third el)))))

	    ;; Right line: I can calculate the x and theta
	    ((eq my-right (cadar el))
	     (setf x (- half-pitch-length 
			(abs (* (second el) (deg-sin (third el))))))
	     (setf theta (if (> (third el) 0)
			     (- 90 (third el))
			   (- -90 (third el)))))

	    ;; Top line: I can calculate the y and theta
	    ((eq my-top (cadar el))
	     (setf y (+ (- 0 half-pitch-width)
			(abs (* (second el) (deg-sin (third el))))))
	     (setf theta (if (> (third el) 0)
			     (- 0 (third el))
			   (- -180 (third el)))))

	    ;; Bottom line: I can calculate the y and theta
	    ((eq my-bottom (cadar el))
	     (setf y (- half-pitch-width
			(abs (* (second el) (deg-sin (third el))))))
	     (setf theta (if (< (third el) 0)
			     (- 0 (third el))
			   (- 180 (third el))))))))

	;; if I've found a line I can try to calculate the other coordinate
	;; To do this I must use a flag
	(when theta

	  ;; foreach object in the percept...
	  (dolist
	   (el (cddr p))

	   ;; if I already know the three params there is nothing to do
	   (unless (and x y theta)

	     ;; (FLAG ...)
	     (when (eq 'flag (caar el))
	       (setf gamma (+ theta (third el)))
	       (cond
	      ;;; See the end of the file for a backup of
	      ;;; source code that was here and uses other
	      ;;; objects to deduce the player position

		;; flags with two letters needed
		((= (length (car el)) 3)
		 (cond

		  ;; (FLAG L ...)
		  ((eq my-left (second (car el)))

		   ;; (FLAG L T)
		   (when (eq my-top (third (car el)))
		     (if x
			 (setf y
			       (- (second flag-l-t-pos)
				  (* (second el) (deg-sin gamma))))
		       (setf x
			     (- (first flag-l-t-pos)
				(* (second el) (deg-cos gamma))))))

		   ;; (FLAG L B)
		   (when (eq my-bottom (third (car el))) ; (flag l b)
		     (if x
			 (setf y
			       (- (second flag-l-b-pos)
				  (* (second el) (deg-sin gamma))))
		       (setf x
			     (- (first flag-l-b-pos)
				(* (second el) (deg-cos gamma)))))))
		
		  ;; (FLAG C ...)
		  ((eq 'c (second (car el)))

		   ;; (FLAG C T)
		   (when (eq my-top (third (car el)))
		     (if x
			 (setf y
			       (- (second flag-c-t-pos)
				  (* (second el) (deg-sin gamma))))
		       (setf x
			     (- (first flag-c-t-pos)
				(* (second el) (deg-cos gamma))))))

		   ;; (FLAG C B)
		   (when (eq my-bottom (third (car el)))
		     (if x
			 (setf y
			       (- (second flag-c-b-pos)
				  (* (second el) (deg-sin gamma))))
		       (setf x
			     (- (first flag-c-b-pos)
				(* (second el) (deg-cos gamma)))))))

		  ;; (FLAG R ...)
		  ((eq my-right (second (car el)))
		   
		   ;; (FLAG R T)
		   (when (eq 't (third (car el))) ; (flag r t)
		     (if x
			 (setf y
			       (- (second flag-r-t-pos)
				  (* (second el) (deg-sin gamma))))
		       (setf x
			     (- (first flag-r-t-pos)
				(* (second el) (deg-cos gamma))))))
		 
		   ;; (FLAG R B)
		   (when (eq my-bottom (third (car el))) ; (flag r b)
		     (if x
			 (setf y
			       (- (second flag-r-b-pos)
				  (* (second el) (deg-sin gamma))))
		       (setf x
			     (- (first flag-r-b-pos)
				(* (second el) (deg-cos gamma)))))))
		  ;; Otherwise
		  (t nil))))))))

	;; update every field known
	(when x (setf (pos-x (agent-kb-player-pos kb)) x))
	(when y (setf (pos-y (agent-kb-player-pos kb)) y))
	(when theta (setf (pos-dir (agent-kb-player-pos kb)) theta))))))


(defun calc-obj-pos (kb obj-p dest)
  "This method calculate the absolute position of an object
   given the current position of the player. Is a lot tricky
   to test all situations to verify that it work."

;; The following turnaround (variables tmp) is needed to bypass
;; a bug of generation of ECL. At runtime a Segmentation
;; Violation is raised. Testing  has shown that when the
;; addend are evaluated in the same function call the
;; exception is raised. Bug verified on Linux RedHat 5.2
;; deg-cos is a macro	
;    (+ (pos-x (player-pos kb))
;       (* rad (deg-cos gamma)))
  (let ((gamma (+ (third obj-p) (pos-dir (agent-kb-player-pos kb))))
	(omega (fifth obj-p))
	(rad (second obj-p)))
    (setf (obj-info-full dest) (cdr obj-p))
    (let ((tmpx (pos-x (agent-kb-player-pos kb)))
	  (tmpcos (* rad (deg-cos gamma)))
	  (tmpy (pos-y (agent-kb-player-pos kb)))
	  (tmpsin (* rad (deg-sin gamma))))

      (setf (obj-info-x dest) (+ tmpx tmpcos)
	    (obj-info-y dest) (+ tmpy tmpsin)
	    (obj-info-dir dest) (if omega omega 0)
	    (obj-info-v dest) (fourth obj-p)))
    ))

;;;; Code backup 
#|	      ((eq 'p (second (car el))) ; Penalty
	       (cond
		((equalp '(l t) (cddar el)) ; (flag p l t)
		 (if x
		     (setf y
			   (+ (second flag-p-l-t-pos)
			      (* (second el) (deg-sin gamma))))
		   (setf x
		   (+ (first flag-p-l-t-pos)
		      (* (second el) (deg-cos gamma))))))
		((equalp '(l c) (cddar el)) ; (flag p l c)
		 (if x
		     (setf y
			   (+ (second flag-p-l-c-pos)
			      (* (second el) (deg-sin gamma))))
		   (setf x
		   (+ (first flag-p-l-c-pos)
		      (* (second el) (deg-cos gamma))))))
		((equalp '(l b) (cddar el))
		 (if x
		     (setf y
			   (+ (second flag-p-l-b-pos)
			      (* (second el) (deg-sin gamma))))
		   (setf x
		   (+ (first flag-p-l-b-pos)
		      (* (second el) (deg-cos gamma))))))
		((equalp '(r t) (cddar el))
		 (if x
		     (setf y
			   (+ (second flag-p-r-t-pos)
			      (* (second el) (deg-sin gamma))))
		   (setf x
		   (+ (first flag-p-r-t-pos)
		      (* (second el) (deg-cos gamma))))))
		((equalp '(r c) (cddar el))
		 (if x
		     (setf y
			   (+ (second flag-p-r-c-pos)
			      (* (second el) (deg-sin gamma))))
		   (setf x
		   (+ (first flag-p-r-c-pos)
		      (* (second el) (deg-cos gamma))))))
		((equalp '(r b) (cddar el))
		 (if x
		     (setf y
			   (+ (second flag-p-r-b-pos)
			      (* (second el) (deg-sin gamma))))
		   (setf x
		   (+ (first flag-p-r-b-pos)
		      (* (second el) (deg-cos gamma))))))
		(t nil))) |#
