;;; @author Antonio Cisternino
;;; @version 0.1
;;;
;;; Purpose: In this file are defined the basic actions
;;; available to an agent. This is the inferior level of
;;; the layers chain and is the real executor of the
;;; actions obtained using the hierarchical planning.
;;;
;;; Changes:
;;;
;;; 16/1/1999: The layer has been expressed in the new language
;;;            for defining plans. See macros.lsp.
;;;
;;; 11/1/1999: The function turn-ball was added.
;;;
;;; 3/1/1999: Introduced the first function that returns multiple actions
;;;           using the release 2.0 of the libraries.

(eval-when (compile eval)
 (plan-file "basic"))

(defupdate
  "This function is very important: if we hear a referee message
   we must replan according to the new state of the game."
  (if (and (eql (agent-kb-last-percept-type kb) 'REFEREE)
	   (not (agent-kb-last-message-read kb)))
      (progn
	(message-read kb)
	'UP)
    nil))

(defactions
  (action-list turn)
  (action-list turn-positive)
  (action-list dash)
  (action-list kick)
  (action-list say)
  (action-list sleep)
  (action-list catch)
  (action-list move)
  (action-list bypass)
  (action-list ndash)
  (action-list turn-ball))

(defaction turn (moment)
  "Usage: turn Moment."
  (if (< (abs moment) 2)
      nil
    `(turn ,moment)))

(defaction turn-positive (moment)
  "Usage: turn Moment."
  ;; Used to avoid problems in reasoning: think positive!
  (change-dir kb moment)
  (if (< (abs moment) 2)
      nil
    `(turn ,moment)))

(defaction dash (power)
  "Usage: dash Power."
  `(dash ,power))

(defaction kick (power dir)
  "Usage: kick Power, Direction."
  (if (zerop dir)
      `((kick 5 ,dir) (kick ,power ,dir))
    `(kick ,power ,dir)))

(defaction say (msg)
  "Usage: say Message."
  `(say ,msg))

(defaction sleep ()
  "Usage: sleep."
  nil)
;  '(dash 0))

(defaction catch (dir)
  "Usage: catch dir."
  `(catch ,dir))

(defaction move (x y)
  "Usage: move x, y."
  `(move ,x ,y))

(defaction bypass ()
  "Usage: bypass l, args. Bypasses the args to the interpreter"
  args)

(defaction ndash (speed times)
  "Usage: ndash speed, times. Execute 'times' dash commands."
  (let ((ret nil))
    (dotimes (i times)
	     (setf ret (append ret `((dash ,speed)))))
    ret))

(defaction turn-ball (direction)
  "Usage: turn-ball [L|R]"
  (rc-turn-trash-off)
  (let ((dir (if (eq direction 'L) '(-80 . -40) '(80 . 40))))
    `((kick 20 ,(car dir)) (turn ,(cdr dir))
      (kick 30 ,(car dir)) (turn ,(cdr dir)) (dash 20))))

