;;; @author Antonio Cisternino
;;; @version 0.1
;;;
;;; Purpose: In this file are defined the basic actions
;;; available to an agent. This is the inferior level of
;;; the layers chain and is the real executor of the
;;; actions obtained using the hierarchical planning.

(eval-when (compile eval)
 (plan-file "goalie-base"))

(def-empty-update)

(defactions
  (action-list catch-ball)
  (action-list follow-action)
  (action-list in-place)
  (action-list launch-ball))
      
(defaction catch-ball ()
  "Usage: catch."
  (let ((ball (last-see-ball kb)))
    (if (< (distance kb (obj-info-x ball) (obj-info-y ball)) 
	   *catch-dist*)
	`((catch ,(agent-kb-last-ball-dir kb)))
      (if (can-catch? kb)
	  `((go-ball ,(very-fast)))
	'((sleep))))))

(defvar *pend* 1)

(defaction follow-action ()
  (let ((ball (last-see-ball kb))
	(dir (pos-dir (agent-kb-player-pos kb))))

    (if ball
	`((turn ,(dir-ball kb)))
      (cond
       ((or (< dir -10) (> dir 10)) 
	(setf *pend* (* *pend* -1))
	`((turn ,(- 0 dir)) (sleep) (sleep)))
       (t
	`((turn ,(* *pend* (- 90 dir))) (sleep) (sleep)))))))

(defaction in-place ()
  (if (not (goalie-in-place? kb))
      `((goto ,*goalie-x* ,*goalie-y* 100) (see-ball))
    '((sleep))))

;; Rivedi
(defaction launch-ball ()
  (let* ((cand (pass-candidates kb))
	 (pl (if cand (select-candidate kb cand) nil)))
    (if pl
	`((pass-ball ,pl))
      `((kick 100 0) (sleep) (sleep)))))
