;;; @author Antonio Cisternino
;;; @version 0.1
;;;
;;; Purpose: In this file are defined the basic actions
;;; available to an agent. This is the inferior level of
;;; the layers chain and is the real executor of the
;;; actions obtained using the hierarchical planning.
;;;
;;; Changes:
;;;
;;; 16/1/1999: The file has been rewritten using the new macro language for
;;;            defining plans.
;;;
;;; 9/1/1999: Now the action are represented as lisp style call function

(eval-when (compile eval)
 (plan-file "player-base"))

(defconstant *max-times* 7)
(defconstant *runwb-times* (if (> *max-times* 7)
			       7
			     *max-times*))

(defvar *dir* 0) ; Used in goto plan

(def-empty-update)

(defactions
  (action-list go-ball)
  (action-list see-ball)
  (action-list pass-ball)
  (action-list run-with-ball)
  (action-list kick!)
  (action-list see-enemy-goal)
  (action-list goto)
  (action-list turn-with-ball))

;; INTRODURRE SCHIVA OGGETTI
(defaction go-ball (speed)
  "Usage: go-ball speed."

  ;; First of all I remember where I've seen the ball?
  (let* ((ball (last-see-ball kb))
	 (bd (when ball (dir-ball kb)))
	 (corr 0)
	 (real-speed (when ball (if (and
				     (> (abs bd) 5)
				     (obj-near kb ball))
				    (if (obj-very-near kb ball)
					(truncate (* 0.75 speed))
				      (truncate (* 0.9 speed)))
				  speed))))

    (if ball
	;; Ok! Now turn in direction of the ball and RUN!!!
	(if (or
	     (and (< (abs (obj-info-dir ball)) 2)
		  (< (abs bd) 10))
	     (and (<= (obj-info-dir ball) 0)
		  (>= bd -4)
		  (<= bd 10))
	     (and (>= (obj-info-dir ball) 0)
		  (<= bd 4)
		  (>= bd 10)))
	    `((ndash ,real-speed ,*max-times*))

	  (progn
	    ;; Evita gli ostacoli, puo` essere migliorato (v. costante 30)
	    (when (and (< (abs bd) 5) (> (agent-kb-front-players kb) 0))
	      (if (> (- (agent-kb-left-players kb) (agent-kb-right-players kb)) 0)
		  (incf bd 30)
		(decf bd 30)))

	    ;; Can be improved!
	    (when (see-ball? kb)
	      (setq corr (* (obj-info-dir (agent-kb-ball-pos kb)) (+ 1 *ball-decay*))))

	    `((turn ,(+ bd corr))
	      (ndash ,real-speed ,*max-times*))))
	    
      ;; Acc! I've forget it... I must find the ball!
      `((*while* (not (see-ball? ,kb)) 
		 ((turn 20) (sleep)))))))

(defaction run-with-ball (speed dir allungo)
  "Usage: run-with-ball speed dir allungo. 
   The power of kick is speed * allungo."

  ;; Evita gli ostacoli, puo` essere migliorato (v. costante 30)
  (when (and (< (abs dir) 5) (> (agent-kb-front-players kb) 0))
    (if (> (- (agent-kb-left-players kb) (agent-kb-right-players kb)) 0)
	(incf dir 30)
      (decf dir 30)))

  (let* ((ball (last-see-ball kb))
	 (bdir (when ball (dir-ball kb))))
    (if (can-kick? kb)
	(if (and (< (abs bdir) 5) (< (abs dir) 5))
	    `((kick ,(* allungo speed) ,dir)
	      (ndash ,speed ,*runwb-times*))
	  `((turn ,bdir)
	    (kick ,(* allungo speed) ,dir)
	    (turn ,(- bdir 0))
	    (turn ,dir)
	    (ndash ,speed ,*runwb-times*)))
      '((sleep)))))

(defaction see-enemy-goal ()
  "Usage: see-enemy-goal"
  (let* ((dir (dir-x-y kb half-pitch-length 0))
	 (kdir dir)
	 (power 20))

    (when (< dir -45)
      (setf kdir -60
	    dir -45))

    (when (> dir 45)
      (setf kdir 60
	    dir 45))

    (if (can-kick? kb)
	`((*while* (and (can-kick? ,kb) (not (see-enemy-goal? ,kb)))
		   ((turn-ball ,(if (< dir 0) 'L 'R)))))
      `((turn ,dir)))))

(defaction kick! (power type)
  "Usage: kick! power type.
   For the moment the parameter type is ignored."
  (if (and 
       (can-kick? kb)
       (see-enemy-goal? kb))
      `((kick ,power ,(third (agent-kb-enemy-goal kb))))
    '((sleep))))
      
(defaction goto (x y speed)
  "Usage: goto x y speed."
  (if (> (distance kb x y) 3)
#|
      `((turn ,(dir-x-y kb x y))
	(ndash ,speed ,*max-times*)
	(*while* (and (let ((ball (last-see-ball ,kb)))
			(if ball
			    (> (distance ,kb (obj-info-x ball) (obj-info-y ball)) 15)
			  t))
		      (< (abs (dir-x-y ,kb ,x ,y)) 5)
		      (> (distance ,kb ,x ,y) 10))
		 ((ndash ,speed ,*max-times*)))
	(*while* (and (< (abs (dir-x-y ,kb ,x ,y)) 5)
		      (> (distance ,kb ,x ,y) 10))
		 ((ndash ,(slow) ,*max-times*))))
|#
      (let ((dir (dir-x-y kb x y)))
	;; Evita gli ostacoli, puo` essere migliorato (v. costante 30)
	(when (and (< (abs dir) 5) (> (agent-kb-front-players kb) 0))
	  (if (> (- (agent-kb-left-players kb) (agent-kb-right-players kb)) 0)
	      (incf dir 30)
	    (decf dir 30)))

	(if (> (abs dir) 5)
	    `((turn ,dir)
	      (ndash ,speed ,*max-times*))
	  `((ndash ,speed ,*max-times*))))
    '((sleep))))

(defaction see-ball ()
  "Usage: see-ball"
  
  ;; First of all I remember where I've seen the ball?
  (let ((ball (last-see-ball kb)))
    (if ball
	`((turn ,(dir-ball kb)))
      `((turn ,(random 90))))))

;; FARE LA FUNZIONE PER REGOLARE L'INTENSITA' DEL PASSAGGio!
(defaction pass-ball (player)
  ;; Can I kick :?
  (if (can-kick? kb)
      (let ((dist (distance kb (obj-info-x player) (obj-info-y player))))
	(if (< (abs (obj-info-dir player)) 45)
	    ;; OK, I can kick! |B<
	    `((kick ,(min
		      100
;;		      (max
;;		      50
		       (* 100 (sqrt (/ dist *pass-ideal-distance*))))
		    ,(dir-x-y kb (obj-info-x player) (obj-info-y player))))

	  ;; I must turn with ball!
	  (if (< (dir-x-y kb (obj-info-x player) (obj-info-y player)) 0)
	      '((turn-ball L))
	    '((turn-ball R)))))
    '((sleep))))

(defaction turn-with-ball (moment)
  "Usage: turn-with-ball moment"
  (let ((final (+ moment (pos-dir (agent-kb-player-pos kb)))))
    (if (> final 180)
	(setf final (- final 360))
      (when (< final -180)
	(setf final (+ final 360))))

    `((*while* (and
		(can-kick? ,kb)
		(< (abs (- (pos-dir ,(agent-kb-player-pos kb)) ,final)) 5))
	       ((turn-ball ,(if (< moment 0) 'L 'R)))))))
