;;;; 
;;;; Last update: 9/1/1999
;;;; @author Antonio Cisternino
;;;; @version 0.2
;;;;
;;;; Purpose: here there are the predicates availables
;;;;          for writing plans
;;;;
;;;; Changes:
;;;;
;;;; 11/4/1999: Added the can-go-ball? predicate.
;;;;
;;;; 25/2/1999: Added the in-zone predicate
;;;;
;;;; 9/1/1999: Changed the predicate enemy-goal: now handle a flat
;;;;           list.
;;;;
;;;; 9/1/1999: functions have been moved to the file functions.lsp
;;;;
;;;; 2/1/1999: added a lot of comments

(defun referee-msg-is? (kb sym)
  "Return T if the last message of the referee has type sym, nil ow."
  (eq (car (agent-kb-last-referee-msg kb)) sym))

;;; Say if the game it is not stopped
;;; I suppose that will be deleted
(defun can-move? (kb)
  (if (agent-kb-last-referee-msg kb)
      (not (member (car (agent-kb-last-referee-msg kb))
		   '(before-kick-off half-time time-up)))
    nil))

(defun see-ball? (kb)
  "Returns the information related to the ball if they are updated."
  (when (agent-kb-last-ball-dir kb)
      (agent-kb-ball-pos kb)))

(defun can-kick? (kb)
  "Decide if the player can kick the ball..."
  (if (agent-kb-last-ball-dist kb)
    (< (agent-kb-last-ball-dist kb) *kick-dist*)))

(defun can-catch? (kb)
  "Decide if the goalie can catch the ball."
  (let ((ball (last-see-ball kb)))
    (if (and ball (< (rc-distance
		      *start-x* *start-y*
		      (obj-info-x ball) (obj-info-y ball)) *goalie-catch-distance*))
	ball
      nil)))

(defun see-enemy-goal? (kb)
  "Say if the player see the enemy goal."
  (let ((goal (agent-kb-enemy-goal kb)))
    (if (and goal
	     (= (cadr (agent-kb-last-percept kb)) (car goal)))
	t
	nil)))

;;; Note: it is necessary?
(defun see-goal? (p)
  "Check if the player see the goal"
  (find-seen-object 'goal p))

(defun in-zone? (kb zone)
  "Decide if the player is within a zone of the pitch. Parameters are 
expressed in 1/nth of pitch."
  (let ((rx (remap-x (first zone)))
	(ry (remap-y (second zone)))
	(rw (remap-w (third zone)))
	(rh (remap-h (fourth zone)))
	(px (pos-x (agent-kb-player-pos kb)))
	(py (pos-y (agent-kb-player-pos kb))))

    (and (> px rx) (< px (+ rx rw)) (> py ry) (< py (+ ry rh)))))

(defun ball-in-zone? (kb zone)
  "Decide if the player is within a zone of the pitch. Parameters are 
expressed in 1/nth of pitch."
  (let ((ball (last-see-ball kb)))
    (if ball
	(let ((rx (remap-x (first zone)))
	      (ry (remap-y (second zone)))
	      (rw (remap-w (third zone)))
	      (rh (remap-h (fourth zone)))
	      (px (obj-info-x ball))
	      (py (obj-info-y ball)))
	  
	  (and (> px rx) (< px (+ rx rw)) (> py ry) (< py (+ ry rh))))
      nil)))

(defun ball-in-attack? (kb)
  (let ((ball (last-see-ball kb)))
    (if ball
	(if (> (obj-info-x ball) 0)
	    t
	  nil)
      'undef)))

(defun goalie-in-place? (kb)
  (if (< (distance kb *goalie-x* *goalie-y*) 2)
      t
    nil))

(defun have-ball? (kb)
  (let ((ball (last-see-ball kb)))
    (if (and ball (< (distance kb (obj-info-x ball) (obj-info-y ball)) 10))
	t
      nil)))

(defun can-go-ball? (kb)
  (when (last-see-ball kb)
	(let ((en (enemies-near-ball kb))
	      (fr (friends-near-ball kb)))
	  (or
;;	   (zerop en) 
	   (< fr 2)
	   (>= en fr)))))

(defun offside? (kb pl)
  (let ((count 0)
	(pl nil))
    (dotimes (i 11)
	     (setf pl (enemy kb (+ i 1)))
	     (when (and pl (> (obj-info-x pl) (pos-x (agent-kb-player-pos kb))))
		   (incf count)))
    (<= count 1)))
